#ifndef StdBasics_hh
#define StdBasics_hh

#include <iostream>
#include <fstream>
#include <assert.h>
#include <vector>
#include <list>
#include <set>
#include <map>
#include <string>
#include <sys/timeb.h>

#include "Utility.hh"

using namespace std;

// Some handy macros
#ifndef NYI
#define NYI { std::cerr<<__FUNCTION__<<" is NYI in "<<__FILE__<<":"<<__LINE__<<std::endl; abort(); }
#endif

#define DFT_APP_ID 1
#define DFT_CTL_ID 1
#define DFT_USE_NAC 0
#define DFT_AUTO_BASELINE 1


// Some default values for tree sampler
#define DFT_MAX_TREE_ITER 0
#define DFT_MAX_TREE_TIME 0
#define DFT_NUM_NODE_EXTRA_FEA 0
#define DFT_MAXNODERAND 0
#define DFT_MAXEDGERAND 0
#define DFT_EDGE_FEA_INTENSITY 1
#define DFT_EDGE_REWARD_DEF 1
#define DFT_USE_LOCAL_REWARD 0
#define DFT_VERBOSITY 0


// Some constants, global variables and functions for PG/RL
#define CWIDTH 12 // Column width for output
#define PRECISION 5 // Precision for output

#define PG_MACHINE_EPS 1e-10

#define DFT_DISCOUNT 0.6
#define DFT_STEP_SIZE 0.001
#define DFT_STEPS_PER_EPOCH 1000
#define DFT_MAX_RL_STEP 0  // 0 is no max
#define DFT_MAX_RL_TIME 0   // 0 is no max
#define DFT_NN_LAYERS 2 // includes input dimension and output dimension
#define DFT_PARAM_FILE "crf.params"
#define DFT_APPROX_OUTPUTS 1
#define DFT_TD_DISCOUNT 0.5

extern double discount, step_size, max_rand_param, td_discount;
extern int steps_per_epoch, max_rl_steps, max_rl_time, nn_layers, approx_outputs;
extern char *param_file;

extern void parse_para(int *argc, char **argv);


#endif
