#include "PGBasics.hh"
#include "CRFController.hh"
#include "Sampler.hh"
#include <boost/numeric/ublas/blas.hpp>
//#include "CyclicPolicyBias.hh"

using namespace tree_mcmc;

namespace libpg {

/**
 * @param  parameterized distribution we'll be using today
 */

CRFController::CRFController(int *argc, char** argv) 
{		
	char flag_name[100];
	
	// To initialize the network, put all parameters in argc, argv !

	// Finally, initialize tree MCMC sampler
	Utility::get_arg(&app_id, argc, argv, "-app", 1, 0);
	switch(app_id)
	{
	case 1:
		crf_engine = new CRFToyTraffic(argc, argv);
		break;
	case 2:
		crf_engine = new CRFSensorChain(argc, argv);
		break;
	case 3:
		crf_engine = new CRFSensorCircle(argc, argv);
		break;
	case 4:
		crf_engine = new CRFOffset(argc, argv);
		break;
	}
	
	crf_engine->init_tree_sampler(argc, argv);
	
// 	filename = NULL;
// 	if(Utility::get_arg(&filename, argc, argv, "-con", 0, NULL))		// Continue file name (for mixture only)
// 	{
// 		continue_fea_para(filename);
// 		delete []filename;
// 	}

	printf("In CRFController, ");
	step_size[1] = 0.001;
	for (int i = 1; i <= 3; i ++)
	{
		sprintf(flag_name, "-ss%d", i);
		Utility::get_arg(&step_size[i], argc, argv, flag_name, 0, step_size[1]);
		ori_step_size[i] = step_size[i];
		printf("%s = %g,\t", flag_name, step_size[i]);
	}
	printf("\n");

	crf_engine->set_step_size(step_size);
}


/**
 * Where the real coding is done to implement softmax
 */
void CRFController::getAction(Observation& obs, Vector& action, bool computeGrad) 
{
//	crf_engine->self_test();


	crf_engine->fill_obs(obs.getFeatures());

	if (crf_engine->get_cur_step() > 0)
		crf_engine->load_last_action();
	
	crf_engine->incre_step();
	crf_engine->set_potentials();

	if (computeGrad)
	{
		crf_engine->infer_graph();
		crf_engine->cal_fea_expectation();
		crf_engine->fill_trace();
	}
 	else
 		crf_engine->sample_graph();

	crf_engine->load_sample_action(action);

	crf_engine->cal_diagnosis();
}


/**
 * Return the most probable action for evaluation purposes only.
 * @param  Observation vector
 * @param  empty vector to load with the most probably action
 */
void CRFController::getMostProbAction(Observation& obs, Vector& action) { 

//     assert(action.size() == 1);
//     approx->doApprox(obs, dist);
// 
//     // Pick most probable action.
//     action[0] = 0;
// 
//     for (size_t i = 1; i < dist.size(); i++) {
//     	if (dist[i] > dist[(int)action[0]]) action[0] = i;
//     }
       
}


/**
 * Just return the input dimensionality of each agent
 * Note: the returned number has NOT been multiplied by the number of agents
 */
int CRFController::getInputDim() 
{
    return crf_engine->get_input_dim();
}


/**
 * The remaining methods are previously just passing into equiv Approximator
 * functions.  Now we do it directly at Controller level, as CRF is a global
 * Controller.
 */
void CRFController::discountTrace() 
{
	crf_engine->discount_trace(discount);
}


void CRFController::setDiscount(double discount) 
{
    this->discount = discount;
}


void CRFController::accumulateGrad(Vector& rewards, Observation& newObs) 
{    
//	assert((int)rewards.size() == global_num_node);

    NYI
}


void CRFController::resetGrad() 
{	
	NYI
}


void CRFController::instantStep(Vector& cur_rewards) 
{	
	crf_engine->instant_step(cur_rewards);
}


void CRFController::computeDirection(int steps) 
{
    NYI
}


void CRFController::batchStep() 
{ 
	crf_engine->batch_step();
}


void CRFController::setStepSize(double _step_size) 
{ 
	return;
// 	if (fabs(ori_step_size[1]) > 1e-100)
// 	{
// 		step_size[1] = _step_size;
// 		for (int i = 1; i <= 3; i ++)
// 			step_size[2] = _step_size * ori_step_size[i] / ori_step_size[1];
// 	}
// 
// 	crf_engine->set_step_size(step_size);
}


void CRFController::resetTrace() 
{ 
	crf_engine->reset_trace();
}


void CRFController::resetParams() 
{
	crf_engine->randomize_node_fea_para();
	crf_engine->randomize_edge_fea_para();	
}


void CRFController::randomizeParams(double maxRand) 
{	
	double rec1, rec2;

	rec1 = maxEdgeRand;
	rec2 = maxNodeRand;

	maxEdgeRand = maxNodeRand = maxRand;
	crf_engine->randomize_node_fea_para(maxRand);
	crf_engine->randomize_edge_fea_para(maxRand);

	maxEdgeRand = rec1;
	maxNodeRand = rec2;
}


double CRFController::getMaxParam() 
{
	return crf_engine->get_max_param();	
}


int CRFController::getNumParams() {
    
	return crf_engine->get_num_params();
}


void CRFController::write(std::ostream& o) 
{
	crf_engine->graph_para_write(o);
}


void CRFController::read(istream& o) 
{
	crf_engine->graph_para_read(o);
}


void CRFController::reduce(Vector& v, Approximator::StatsEnum s) 
{
	if (Approximator::PARAMS != 0)
	{
		cout<<"Approximator::PARAMS != 0"<<endl;
		exit(1);
	}
		
	if (Approximator::GRADS != 1)
	{
		cout<<"Approximator::GRADS != 1"<<endl;
		exit(1);
	}
		
	if (Approximator::TRACES != 2)
	{
		cout<<"Approximator::TRACES != 2"<<endl;
		exit(1);
	}

	crf_engine->reduce(v, s);
}


void CRFController::scatter(Vector& v, Approximator::StatsEnum s) 
{
	if (Approximator::PARAMS != 0)
	{
		cout<<"Approximator::PARAMS != 0"<<endl;
		exit(1);
	}
		
	if (Approximator::GRADS != 1)
	{
		cout<<"Approximator::GRADS != 1"<<endl;
		exit(1);
	}
		
	if (Approximator::TRACES != 2)
	{
		cout<<"Approximator::TRACES != 2"<<endl;
		exit(1);
	}

	crf_engine->scatter(v, s);
}


// void CRFController::continue_fea_para(char *filename)
// {
// 	int cur_item, cur_fea, useless1, useless2;
// 	double ***pd3, **pd2, *pd1, max_para;
// 	FILE *p_input;
// 
// 	max_para = getMaxParam();
// 	pd3 = node_fea_para + 1;
// 	
// 	p_input = fopen(filename, "r");
// 	for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
// 	{
// 		pd2 = *(pd3 ++) + 1;
// 
// 		fscanf(p_input, "[%d,%d]((%lf),(%lf))\n\n", &useless1, &useless2, *pd2 + 1, *pd2 + 2);
// 		pd2 ++;
// 		*(*pd2 + 1) = *(*pd2 + 2) = 0;
// 	}
// 	
// 	pd2 = edge_fea_para + 1;
// 	for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)
// 	{
// 		pd1 = *(pd2 ++) + 1;
// 		for (cur_fea = 1; cur_fea <= num_edge_fea; cur_fea ++)
// 			*(pd1 ++) = max_para * (rand() - RAND_MAX / 2) / RAND_MAX;
// 	}
// }


char * CRFController::get_diagnosis(int argc, char ** argv)
{
	return crf_engine->get_diagnosis(argc, argv);
}


}


