#ifndef NACTransform_hh
#define NACTransform_hh

// $Id: NACTransform.hh 29 2006-12-19 02:16:20Z daa $

namespace libpg {
class NACTransform : public TransformController {


protected:

    int nacDimension;		// Size of basis vectors plus grads

    Observation lastObs;	// Copy of last observation
    double lastReward;		// Copy of last reward
    Vector vec_lastReward;	// vector version of lastReward
    bool instantStepRequired; // Are we waiting for the next obs to complete step?

    double discount;		// Local copy of this. Always the same as controller discount

    Matrix AInverse;
    Vector nacTrace;		// Trace variable for features				: e_t in the pseudo-code (step 6)
    Vector nacGrad;			// accumulated reward times obsTrace		: \delta_t in the pseudo-code (step 9, 16)
    Vector nacDirection;	// computed search direction				: [dt, vt] in the pseudo-code (step 16)
    Vector nacTD;			// Temporal difference in basis, inc. grads,: w_t in the pseudo-code (step 11)
    Vector nacParamUpdate;	// Search direction proposed by NAC.		: d_t in the pseudo-code (step 17)
    Matrix I;				// identity matrix shortcut
    
    double tdDiscount;		// Critic discount factor
	double *rewards;
    static bool print_nac_info;

public:
    
    NACTransform(Controller* controller, double tdDiscount);

    virtual void discountTrace();

    virtual void getAction(Observation& obs, Vector& action, bool computeGrad = true);
    virtual void getMostProbAction(Observation& obs, Vector& action) { NYI }

    virtual void resetTrace();
    virtual void resetGrad();
    virtual void accumulateGrad(double reward, Observation& newObs);
    virtual void instantStep(Vector& reward);
    virtual void delayedInstantStep(double reward, Observation& newObs);
    virtual void computeDirection(int steps);
    virtual void batchStep();
    virtual void updateStatistics(double reward, Observation& newObs);
    virtual void scatter(Vector& v, Approximator::StatsEnum s);
    virtual void reduce(Vector& v, Approximator::StatsEnum s);
    virtual void setDiscount(double discount);
	virtual char * get_diagnosis(int argc, char ** argv);
	virtual void buffer_last_action(int * actions)  {controller->buffer_last_action(actions);};
    
	void dist_rwd_updateStatistics(Vector & reward, Observation& newObs);
	void dist_rwd_delayedInstantStep(Vector & reward, Observation& newObs);
};
}
#endif
