#include"PGBasics.hh"
#include"CRFController.hh"
#include"Sampler.hh"
//#include"GridSim.hh" 
#include <boost/numeric/ublas/blas.hpp>


using namespace std;

/**
 * @param  parameterized distribution we'll be using today
 */

CRFController::CRFController(int size_x, int size_y, int obsRow, int actDim, int numEdgeFea, 
							 int numAgent, int numObs, int numAction, int argc, char** argv) {

	this->numAgent = numAgent;			// gridSize * gridSize
	this->numObs = numObs;			// number of observations
	this->numAction = numAction;		// number of actions

//     nodeTrace.resize(numNodeFeat);
//     edgeTrace.resize(numEdgeFeat);

	initializeNetwork(size_x, size_y, obsRow, actDim, numEdgeFea, argc, argv);
}


/**
 * Where the real coding is done to implement softmax
 */
void CRFController::getAction(Observation& obs, Vector& action, bool computeGrad) {

    unsigned int i;

    assert(action.size() == numAgent);

	run_inference();

	for(i = 0; i < numAgent; i ++)
		action[i] = sample_record[i];

//	updateNodePotential();

	// Finally update the gradient, and add to the trace
    if (computeGrad) {
		// now we need to fill in the feature gradients (expectations)
		fill_trace();
    }

}


/**
 * Return the most probable action for evaluation purposes only.
 * @param  Observation vector
 * @param  empty vector to load with the most probably action
 */
void CRFController::getMostProbAction(Observation& obs, Vector& action) { 

//     assert(action.size() == 1);
//     approx->doApprox(obs, dist);
// 
//     // Pick most probable action.
//     action[0] = 0;
// 
//     for (size_t i = 1; i < dist.size(); i++) {
//     	if (dist[i] > dist[(int)action[0]]) action[0] = i;
//     }
       
}


/**
 * Just return the input dimensionality of each agent
 * Note: the returned number has NOT been multiplied by the number of agents
 */
int CRFController::getInputDim() {
    return numObs;
}


/**
 * The remaining methods are traditionally just pass onto equiv Approximator
 * functions.  Now we do it directly at Controller level, as CRF is a global
 * Controller.
 */

void CRFController::discountTrace() {

	unsigned int i, j, k;
	double *** cur_3, ** cur_2, * cur_1;

	//  First, nodes
	cur_3 = node_fea_grad;

	for(i = 0; i < num_nodes; i++)
	{
		cur_2 = *(cur_3++);

		for(j = 0; j < dim_action; j++)
		{			
			cur_1 = *(cur_2++);

			for(k = 0; k < dim_observation; k++)
			{
				*(cur_1++) *= discount;	
				// update node feature parameter here, this is actually node_fea_para[i][j][k]
			}
		}
	}

	cur_2 = edge_fea_grad;

	// Second, horizontal edges
	for(i = 0; i < dim_vertical; i++)
	{
		for(j = 0; j < dim_horizontal - 1; j++)
		{
			cur_1 = *(cur_2++);

			for(k = 0; k < num_edge_fea; k++)
			{
				*(cur_1++) *= discount;		// 	update horizontal edge grads, edge_fea_grad[i*dim_horizontal+j][k]
			}
		}
		cur_2++;
	}

	// Third, vertical edges
	for(i = dim_vertical * dim_horizontal; i < num_edge_max; i++)
	{
		cur_1 = *(cur_2++);

		for(k = 0; k < num_edge_fea; k++)
		{
			*(cur_1++) *= discount;			// update the vertical edge grads, edge_fea_para[i][k]
		}
	}	

}


void CRFController::setDiscount(double discount) {

    this->discount = discount;
}


void CRFController::accumulateGrad(Vector& rewards, Observation& newObs) {
    
	assert(rewards.size() == numAgent);

    NYI
}


void CRFController::resetGrad() {
    NYI
}


void CRFController::instantStep(Vector& rewards) {

    assert(rewards.size() == 1);	// we just accept one global reward
	
	if (fabs(rewards[0] - 0.0) > 1e-5) 
	{
		update_parameters_grid(rewards[0] * stepSize);
	}		
}


void CRFController::computeDirection(int steps) {
    NYI
}


void CRFController::batchStep() { 
    NYI 
}


void CRFController::setStepSize(double stepSize) { 
    this->stepSize = stepSize;
}


void CRFController::resetTrace() { 
	unsigned int i, j, k;
	double *** cur_3, ** cur_2, * cur_1;

	//  First, nodes
	cur_3 = node_fea_grad;

	for(i = 0; i < num_nodes; i++)
	{
		cur_2 = *(cur_3++);

		for(j = 0; j < dim_action; j++)
		{			
			cur_1 = *(cur_2++);

			for(k = 0; k < dim_observation; k++)
			{
				*(cur_1++) = 0;	 // update node feature grads, node_fea_para[i][j][k]
			}
		}
	}

	cur_2 = edge_fea_grad;

	// Second, horizontal edges
	for(i = 0; i < dim_vertical; i++)
	{
		for(j = 0; j < dim_horizontal - 1; j++)
		{
			cur_1 = *(cur_2++);

			for(k = 0; k < num_edge_fea; k++)
			{
				*(cur_1++) = 0;			// 	update horizontal edge grads, edge_fea_grad[i*dim_horizontal+j][k]
			}
		}
		cur_2++;
	}

	// Third, vertical edges
	for(i = dim_vertical * dim_horizontal; i < num_edge_max; i++)
	{
		cur_1 = *(cur_2++);

		for(k = 0; k < num_edge_fea; k++)
		{
			*(cur_1++) = 0;			// update the vertical edge grads, edge_fea_para[i][k]
		}
	}	
}


void CRFController::resetParams() { 

	unsigned int i, j, k;
	double *** cur_3, ** cur_2, * cur_1;

	//  First, nodes
	cur_3 = node_fea_para;

	for(i = 0; i < num_nodes; i++)
	{
		cur_2 = *(cur_3++);

		for(j = 0; j < dim_action; j++)
		{			
			cur_1 = *(cur_2++);

			for(k = 0; k < dim_observation; k++)
			{
				*(cur_1++) = 0;	 // update node feature grads, node_fea_para[i][j][k]
			}
		}
	}

	cur_2 = edge_fea_para;

	// Second, horizontal edges
	for(i = 0; i < dim_vertical; i++)
	{
		for(j = 0; j < dim_horizontal - 1; j++)
		{
			cur_1 = *(cur_2++);

			for(k = 0; k < num_edge_fea; k++)
			{
				*(cur_1++) = 0;			// 	update horizontal edge grads, edge_fea_grad[i*dim_horizontal+j][k]
			}
		}
		cur_2++;
	}

	// Third, vertical edges
	for(i = dim_vertical * dim_horizontal; i < num_edge_max; i++)
	{
		cur_1 = *(cur_2++);

		for(k = 0; k < num_edge_fea; k++)
		{
			*(cur_1++) = 0;			// update the vertical edge grads, edge_fea_para[i][k]
		}
	}
}


void CRFController::randomizeParams(double maxRand) {     
// 	UBlasExtras::randomize(nodeParams, maxRand);
// 	UBlasExtras::randomize(edgeParams, maxRand);
}


double CRFController::getMaxParam() {

	unsigned int i, j, k;
	double *** cur_3, ** cur_2, * cur_1, record;

	record = -1;

	//  First, nodes
	cur_3 = node_fea_para;

	for(i = 0; i < num_nodes; i++)
	{
		cur_2 = *(cur_3++);

		for(j = 0; j < dim_action; j++)
		{			
			cur_1 = *(cur_2++);

			for(k = 0; k < dim_observation; k++)
			{
				if(fabs(*cur_1) > record)
					record = fabs(*cur_1);
				cur_1 ++;
			}
		}
	}

	cur_2 = edge_fea_para;

	// Second, horizontal edges
	for(i = 0; i < dim_vertical; i++)
	{
		for(j = 0; j < dim_horizontal - 1; j++)
		{
			cur_1 = *(cur_2++);

			for(k = 0; k < num_edge_fea; k++)
			{
				if(fabs(*cur_1) > record)
					record = fabs(*cur_1);
				cur_1 ++;
			}
		}
		cur_2++;
	}

	// Third, vertical edges
	for(i = dim_vertical * dim_horizontal; i < num_edge_max; i++)
	{
		cur_1 = *(cur_2++);

		for(k = 0; k < num_edge_fea; k++)
		{
			if(fabs(*cur_1) > record)
				record = fabs(*cur_1);
			cur_1 ++;
		}
	}

	return record;
}


int CRFController::getNumParams() {
    
	return num_edge_fea * ( 2 * dim_horizontal * dim_action - dim_horizontal - dim_vertical)
			+ dim_action * dim_observation * dim_horizontal * dim_vertical;
}


void CRFController::write(ostream& o) {

	unsigned int i, j, k;
	double *** cur_3, ** cur_2, * cur_1;

	//  First, nodes
	cur_3 = node_fea_para;

	for(i = 0; i < num_nodes; i++)
	{
		cur_2 = *(cur_3++);

		for(j = 0; j < dim_action; j++)
		{			
			cur_1 = *(cur_2++);

			for(k = 0; k < dim_observation; k++)
			{
				o<<*(cur_1++);	 // update node feature grads, node_fea_para[i][j][k]
			}
		}
	}

	cur_2 = edge_fea_para;

	// Second, horizontal edges
	for(i = 0; i < dim_vertical; i++)
	{
		for(j = 0; j < dim_horizontal - 1; j++)
		{
			cur_1 = *(cur_2++);

			for(k = 0; k < num_edge_fea; k++)
			{
				o<<*(cur_1++);			// 	update horizontal edge grads, edge_fea_grad[i*dim_horizontal+j][k]
			}
		}
		cur_2++;
	}

	// Third, vertical edges
	for(i = dim_vertical * dim_horizontal; i < num_edge_max; i++)
	{
		cur_1 = *(cur_2++);

		for(k = 0; k < num_edge_fea; k++)
		{
			o<<*(cur_1++);			// update the vertical edge grads, edge_fea_para[i][k]
		}
	}
}


void CRFController::read(istream& o) {

	unsigned int i, j, k;
	double *** cur_3, ** cur_2, * cur_1;

	//  First, nodes
	cur_3 = node_fea_para;

	for(i = 0; i < num_nodes; i++)
	{
		cur_2 = *(cur_3++);

		for(j = 0; j < dim_action; j++)
		{			
			cur_1 = *(cur_2++);

			for(k = 0; k < dim_observation; k++)
			{
				o>>*(cur_1++);	 // update node feature grads, node_fea_para[i][j][k]
			}
		}
	}

	cur_2 = edge_fea_para;

	// Second, horizontal edges
	for(i = 0; i < dim_vertical; i++)
	{
		for(j = 0; j < dim_horizontal - 1; j++)
		{
			cur_1 = *(cur_2++);

			for(k = 0; k < num_edge_fea; k++)
			{
				o>>*(cur_1++);			// 	update horizontal edge grads, edge_fea_grad[i*dim_horizontal+j][k]
			}
		}
		cur_2++;
	}

	// Third, vertical edges
	for(i = dim_vertical * dim_horizontal; i < num_edge_max; i++)
	{
		cur_1 = *(cur_2++);

		for(k = 0; k < num_edge_fea; k++)
		{
			o>>*(cur_1++);			// update the vertical edge grads, edge_fea_para[i][k]
		}
	}
}


void CRFController::reduce(Vector& v, Approximator::StatsEnum s) {

//    assert(v.size() >= (size_t)(parameters));

// 	if (s==PARAMS) {
// 		project(v, range(0, numEdgeFeat)) += edgeParams;
// 		project(v, range(numEdgeFeat, parameters)) += nodeParams;
// 	}
// 	else  
		NYI;

}


void CRFController::scatter(Vector& v, Approximator::StatsEnum s) {

//    assert(v.size()==(size_t)parameters);

// 	if (s==PARAMS) {
// 		edgeParams.clear();
// 		edgeParams +=  project(v, range(0, numEdgeFeat));
// 
// 		nodeParams.clear();
// 		nodeParams +=  project(v, range(numEdgeFeat, parameters));
// 	}
// 	else  
		NYI;

}


/**
 * set up the network topology and node potentials
 * call PGM
*/
void CRFController::initializeNetwork(int size_x, int size_y, int obsRow, int actDim, int numEdgeFea, int argc, char** argv) {

	initialize_crf_parameters(size_x, size_y, obsRow, actDim, numEdgeFea);

	randomize_parameters_grid(argc, argv);
	
	prepare_string_input(input_string, argc, argv);

	cout<<input_string.str();	
}



/**
 * set up the network topology and node potentials
 * call PGM
*/
void CRFController::run_inference() {

	Experiment *main_experiment;
	
	main_experiment = new Experiment;
	
	input_string.clear();
	input_string.seekg(0, ios::beg);	

	if (!main_experiment->set_up_model_from_string(input_string))
	{
		cout << "A fatal error occurred while parsing file. Aborting..." << endl;
		exit(1);
	}
	
//	print_edge_potential();
//	main_experiment->show_edges();

	main_experiment->get_inference();

//	print_edge_marginals();

	cal_fea_expectations_grid();

//	main_experiment.export_results_to_file( argv[2]);
	
	delete main_experiment;

}


void CRFController::fill_trace()
{
	unsigned int i, j, k, cur_node;
	double *** cur_3, ** cur_2, * cur_1, ***cur_3_other, ** cur_2_other, * cur_1_other;

	//  First, nodes
	cur_3 = node_fea_grad;
	cur_3_other = node_fea_exp;

	for(i = 0; i < num_nodes; i++)
	{
		cur_2 = *(cur_3++);
		cur_2_other = *(cur_3_other++);

		for(j = 0; j < dim_action; j++)
		{
			cur_1 = *(cur_2++);
			cur_1_other = *(cur_2_other++);

			if(sample_record[i] != j)	// did not sample the action j
			{
				for(k = 0; k < dim_observation; k++)
				{
					*(cur_1++) -= *(cur_1_other++);
				}
			}
			else
			{
				for(k = 0; k < dim_observation; k++)
				{
					*(cur_1++) += cur_observation[i][k] - *(cur_1_other++);	
				}
			}
		}
	}

	cur_2 = edge_fea_grad;
	cur_2_other = edge_fea_exp;
	cur_node = 0;

	// Second, horizontal edges
	for(i = 0; i < dim_vertical; i++)
	{
		for(j = 0; j < dim_horizontal - 1; j++)
		{
			cur_1 = *(cur_2++);
			cur_1_other = *(cur_2_other++);

			if(sample_record[cur_node] == sample_record[cur_node + 1])
			{
				for(k = 0; k < num_edge_fea; k++)
				{
					*(cur_1++) += 1 - *(cur_1_other++);
				}
			}
			else
			{
				for(k = 0; k < num_edge_fea; k++)
				{
					*(cur_1++) -= *(cur_1_other++);
				}
			}

			cur_node++;
		}

		cur_2++;		cur_2_other++;		cur_node++;
	}

	// Third, vertical edges
	cur_node = 0;
	for(i = dim_vertical * dim_horizontal; i < num_edge_max; i++, cur_node++)
	{
		cur_1 = *(cur_2++);
		cur_1_other = *(cur_2_other++);

		if(sample_record[cur_node] == sample_record[cur_node + dim_horizontal])
		{
			for(k = 0; k < num_edge_fea; k++)
			{
				*(cur_1++) += 1 - *(cur_1_other++);
			}
		}
		else
		{
			for(k = 0; k < num_edge_fea; k++)
			{
				*(cur_1++) -= *(cur_1_other++);
			}
		}
	}
}

