#ifndef GridSim_hh
#define GridSim_hh
// $Id$

/**
 * Container for any RL problem we care to define. We simply have to
 * define the dimensionality of  actions, a reward function, a way to get
 * observations, and a way to do actions.
 */

using namespace libpg;

class GridSim : public Simulator {

private:

    int num_row, num_col;
    int nodes;
    int bufferSize;
	int ctl_id;
	int count_act1, count_act2;

    // Probability with which a packet arrives at any node
    double packetGenProb;

    // Store the current buffer of packets waiting to get through
    int * row_queue;
    int * col_queue;

    typedef enum { VERTICAL=0, HORIZONTAL=1 } NodeDirection;

    int nextReward;

public:

    GridSim(int *argc, char ** argv);

    virtual ~GridSim();

    /**
     * Get the number of rows in observation matrix. Generally the
     * length of the observation vector.
     */
    virtual int getObsRows();

    /**
     * Get the number of columns in the observation matrix Typically
     * use multiple columns to provide a different observation to each
     * agent, so the number of columns is the same as getAgents()
     */
    virtual int getObsCols();

    /**
     * Get the total number of agents. Usually 1.
     */
    virtual int getAgents();

    /**
     * Get the total action dimensionality. Often, but not
     * necessarily, the same as the number of agents.
     */
    virtual int getActionDim();

    /**
     * Get the dimensionality of the reward vector. Often, but not
     * necessarily, the same as the number of agents.
     */
    virtual int getRewardDim();

    /**
     * Get the vertical size of the Grid.
     */
    int getGrid_numRow();
	
    /**
     * Get the horizontal size of the Grid.
     */
    int getGrid_numCol();

	
	/**
     * Just return the reward associated with the most recent state
     * achieved.  
     * @param rewards vector of rewards. The vector has
     * getAgents() dimensionality In non-local reward mode this vector
     * will just be averaged.
     */
    virtual void getReward(Vector& rewards);

    virtual void getObservation(Observation& obs);


    /**
     * Do action issued by a controller @return  vector of actions,
     * typically just a scalar which is cast to an int, but it there
     * are multiple agents it will be a vector of actions for each
     * agent.
     * @param action vector that describes the action to perform.
     */
    virtual int doAction(Vector& action);

	virtual double get_max_reward();

	virtual char * get_diagnosis();

};

inline int GridSim::getGrid_numRow() { return num_row; }

inline int GridSim::getGrid_numCol() { return num_col; }

#endif
