#include "PGBasics.hh"

#include "definitions.hh"
#include "Ps.hh"
#include "intersection.hh"
#include "car.hh"  

using namespace std;
using namespace libpg;

int  random2()
{
  int r = random();
  //cerr << r << endl;
  return r;
}

/*
class Car {
public:
  Intersection* current_intersection;
  Intersection::Direction current_incoming_direction;
  Intersection::Direction outgoing_direction;
  Intersection::Turn turn;
  Intersection* next_intersection;
  Intersection::Direction next_incoming_direction;
  int position_on_road;
  int queue_number;
  Intersection* source;
  Intersection* destination;
  int create_time;
  void navigate();
};
*/

Car::Car(Intersection* the_source, Intersection* dest, double time, double the_id):
  source(the_source), destination(dest), create_time(time), id(the_id) {
  if(Ps::pVERBOSE || VERBOSE_CAR) {
    cout << "creating car "<< id << ", source: ";
    source->print_id();
    cout << ", destination: ";
    destination->print_id();
    cout << endl;
  }
  current_intersection = source;
  // the following 2 values are defaults that will be overwritten by the navigate method
  current_incoming_direction = Intersection::North;
  position_on_road = 0;
  wait_time = 0;
  navigate();
  position_on_road = next_intersection->get_road_length
      (next_incoming_direction);
  //next_intersection->created_car(next_incoming_direction, position_on_road); 
  if(Ps::pVERBOSE || VERBOSE_CAR)
    cout << "Car " << id << " created." << endl;
}

void Car::do_departure() {
  if(Ps::pVERBOSE || VERBOSE_CAR) {
    cout << "passing through intersection" << endl;
  }
  current_intersection->car_depart(this, current_incoming_direction, turn, wait_time);
  wait_time = 0;
  position_on_road = next_intersection->get_road_length
    (next_incoming_direction);
  next_intersection->enter_car_on_road(next_incoming_direction, 
				       position_on_road);
  if(Ps::pVERBOSE || VERBOSE_CAR) {
    print_info();
  }
}

void Car::print_info() {
  cout << "Car " << id;
  cout << ", position on road: " << position_on_road;
  if(position_on_road == 0) {
    //cout << ", queue no. " << queue_number;
    cout << ", current in dir " << current_incoming_direction;
    cout <<", turn " << turn;
  }
  cout << ", current intersec: ";
  current_intersection->print_id();
  cout << ", " << "next intersec: ";
  next_intersection->print_id();
  cout << endl;
  cout <<", next incoming direction "<< next_incoming_direction << endl;
}

bool Car::update() {
  // updates the position of the car in the map
  if(Ps::pVERBOSE || VERBOSE_CAR)
    print_info();
  if(position_on_road > 0) { //we are travelling along the road  
    if(!next_intersection->is_road_free
	(next_incoming_direction, position_on_road - 1)) { 
      // road is blocked
      if(Ps::pVERBOSE || VERBOSE_CAR) {
	cout << "cannot move to " << position_on_road - 1 << 
	  ", "; 
	next_intersection->print_id();
	cout << ", dir " << next_incoming_direction << endl;
      }
    }
    else { // move one road segment
      next_intersection->advance_car(next_incoming_direction, position_on_road);
      position_on_road--; 
      if(Ps::pVERBOSE || VERBOSE_CAR)
	cout << "decreasing road position to " << position_on_road << endl;
      if(position_on_road == 0) { // we arrived at intersection
	if(Ps::pVERBOSE || VERBOSE_CAR) {
	  cout << "arrived at intersection ";
	  next_intersection->print_id();
	  cout << endl;
	}
	last_intersection = current_intersection;
	current_intersection = next_intersection;
	current_incoming_direction = next_incoming_direction;
	if(current_intersection == destination) {
	  current_intersection->car_arrived_at_destination
	    (current_incoming_direction); 
	  return true; // arrived: car will be removed from system
	}
	turn = navigate(); // else: calculate turn, set next_intersection etc.
	// check into queue
	int queue_number = current_intersection->
	  car_arrive(this, current_incoming_direction, turn);
	if(Ps::pVERBOSE || VERBOSE_CAR)
	  cout << "current queue position is " << queue_number << endl;
	if(queue_number == 0 && 
	   current_intersection->
	   is_phase_matching(current_incoming_direction, turn) &&
	   !current_intersection->is_blocked
	   (current_incoming_direction, turn) &&
	   !current_intersection->is_maxed_out
	   (current_incoming_direction, turn)) {
	  // we can go if road free
	  if(next_intersection->is_road_free(next_incoming_direction)) {
	    assert(wait_time == 0 && current_intersection->
		   queue_pos_leq_throughput_left(this, current_incoming_direction, turn));
	    do_departure();
	  }
	  else {
	    current_intersection->blocked_queue(current_incoming_direction, turn);
	  }
	}
	// we could not leave
	if(!current_intersection->is_phase_matching(current_incoming_direction, turn)) {
	  if(Ps::pVERBOSE || VERBOSE_CAR) {
	    cout << "sending wait signal " << endl;
	    }
	  current_intersection->car_waiting();
	}	
      }
    }
  }
  else { // we are waiting at intersection
    wait_time++;
    // if wrong phase or full road or waiting number to large, keep waiting
    if(!current_intersection->
       is_phase_matching(current_incoming_direction, turn) ||
       //queue_number >= current_intersection->get_car_throughput_left
       //(current_incoming_direction, turn) ||
       !current_intersection->
       queue_pos_leq_throughput_left(this, current_incoming_direction, turn) ||
       current_intersection->is_blocked
       (current_incoming_direction, turn)){
      if(Ps::pVERBOSE || VERBOSE_CAR) {
	cout << "keep waiting, wait_time is " << wait_time << endl;
	cout << "sending wait signal " << endl;
      }
      current_intersection->car_waiting();
      return false;
    }
    // else
    if(!next_intersection->is_road_free(next_incoming_direction)) {
      current_intersection->blocked_queue(current_incoming_direction, turn);
      if(Ps::pVERBOSE || VERBOSE_CAR)
	cout << "keep waiting, wait_time is " << wait_time << endl;
      //current_intersection->car_waiting();
      return false;
    }
    else{
      if(current_intersection->is_maxed_out(current_incoming_direction, turn)) {
	cout << "problematic car: ";
	print_info();
      }
      assert(!current_intersection->is_maxed_out(current_incoming_direction, turn));
      do_departure();
    }
  }
  return false; 
}

/*
void Car::decrease_queue_number() {
  queue_number--;
  if(Ps::pVERBOSE || VERBOSE_CAR) { 
    cout << "Car " << id;
    cout << " decreased queue_number to " << queue_number << endl; 
  }
}
*/

void plan_path_one_step(Intersection* from, Intersection* to,
			std::vector<Intersection::Direction>& dirs) {
  if(Ps::pVERBOSE || VERBOSE_CAR) {
    cout << "from ";
    from->print_id();
    cout << " to ";
    to->print_id();
  }
  int from_x = from->get_x_coordinate();
  int from_y = from->get_y_coordinate();
  int to_x = to->get_x_coordinate();
  int to_y = to->get_y_coordinate();
  Intersection::Direction first;
  Intersection::Direction second;
  bool deterministic = false; 
  bool first_choice_mandatory = false;
  bool second_choice_mandatory = false;
  // in the future: check intersection whether turn is possible
  if(from_x > to_x && from->has_outgoing(Intersection::West))
    first = Intersection::West;
  else if(from_x < to_x && from->has_outgoing(Intersection::East))
    first = Intersection::East;
  else { // x-es equal or no turn in x-direction possible
    if(from_x != to_x && (Ps::pVERBOSE || VERBOSE_CAR)) {
	cout << "cannot turn  in x-direction at this intersection";
      }
      second_choice_mandatory = true;
      first = Intersection::East; // arbitrary, to suppress warnings
  }
  
  if(from_y > to_y && from->has_outgoing(Intersection::North)) {
    second = Intersection::North;
  }
  else if(from_y < to_y && from->has_outgoing(Intersection::South)) {
    second = Intersection::South;
  }
  else { // y's equal or no turn in y-direction possible
    first_choice_mandatory = true;
     if(from_y != to_y && (Ps::pVERBOSE || VERBOSE_CAR)) {
       cout << "cannot turn  in y-direction at this intersection" << endl;
      }
     if(second_choice_mandatory) {
       cout << "cannot find path to goal!" << endl;
       exit(1);
     }
     second = Intersection::South; // arbitrary, to suppress warnings
  }

  if((deterministic && !second_choice_mandatory) || first_choice_mandatory) {
    assert(from->has_outgoing(first));
    dirs.push_back(first);
  }
  else if(second_choice_mandatory) {
    assert(from->has_outgoing(second));
    dirs.push_back(second);
  }
  else {
    // two possible directions
    assert(from->has_outgoing(first));
    dirs.push_back(first);
    assert(from->has_outgoing(second));
    dirs.push_back(second);
  }
}
  
Intersection::Direction choose_randomly(Intersection::Direction first,
					  Intersection::Direction second) {

  Intersection::Direction return_d;
  int r = random2() % 2;
  if(r)
    return_d = first;
  else
    return_d = second;
    return return_d;
}

Intersection::Turn calculate_turn(Intersection::Direction in,
				  Intersection::Direction out) {
  Intersection::Turn turn;
  if(in == (out + 1) % DIRECTIONS)
    turn = Intersection::Right;
  else if(in == (out + 2) % DIRECTIONS)
    turn = Intersection::Straight;
  else // (in == (out + 3) % DIRECTIONS)
    turn = Intersection::Left;
  return turn;
}

Intersection::Direction Car::decide_next_outgoing_dir
                              (std::vector<Intersection::Direction>& dirs) {
  if(dirs.size() == 1) {
    return dirs.at(0); 
  }
  else {
    assert(dirs.size() == 2);
    if(Ps::pGREEDY_DRIVERS) { // choose direction which is green
      bool first_green = false;
      bool second_green = false;
      if(Ps::pVERBOSE || VERBOSE_CAR) 
	cout << "possible directions: " << dirs.at(0) << " " << dirs.at(1) << endl;
      Intersection::Turn turn1 = calculate_turn(current_incoming_direction,
						dirs.at(0));
      Intersection::Turn turn2 = calculate_turn(current_incoming_direction,
						dirs.at(1));
      Intersection::Direction  incoming_direction1 = (Intersection::Direction) 
	((dirs.at(0) + 2) % DIRECTIONS);
      Intersection::Direction  incoming_direction2 = (Intersection::Direction) 
	((dirs.at(1) + 2) % DIRECTIONS);

      if(current_intersection->
	 is_phase_matching(current_incoming_direction, turn1) &&
	 !current_intersection->is_queue_full(current_incoming_direction, turn1) &&
	 !current_intersection->is_blocked(current_incoming_direction, turn1) &&
	 current_intersection->get_neighbouring_intersection(dirs.at(0))
	 ->is_road_free(incoming_direction1)) {
	  first_green = true;
      }
      if(current_intersection->
	 is_phase_matching(current_incoming_direction, turn2) &&
	 !current_intersection->is_queue_full(current_incoming_direction, turn2) &&
	 !current_intersection->is_blocked(current_incoming_direction, turn2) &&
	 current_intersection->get_neighbouring_intersection(dirs.at(1))
	 ->is_road_free(incoming_direction2)) {
	  second_green = true;
      }
    
      if(first_green && !second_green) {
	if(Ps::pVERBOSE || VERBOSE_CAR) {
	  cout << endl << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!";
	  cout << "direction " << dirs.at(0)<< " is green, I'm taking that!" << endl;
	}
	return dirs.at(0);	
      }
      else if(second_green && !first_green) {
	if(Ps::pVERBOSE || VERBOSE_CAR) {
	  cout << endl << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!";
	  cout << "direction " << dirs.at(1)<< " is green, I'm taking that!" << endl;
	}
	return dirs.at(1);
      }
      // both phases (not) green, look for shorter queue
      if(current_intersection->is_queue_full(current_incoming_direction, turn2) &&
	 !current_intersection->is_queue_full(current_incoming_direction, turn1))
	return dirs.at(0);
      else if(current_intersection->is_queue_full(current_incoming_direction, turn1) &&
	      !current_intersection->is_queue_full(current_incoming_direction, turn2))
	return dirs.at(1);
    }
    // no greedy drivers or both queues equal, decide randomly 
    bool deterministic = false;
    if(!deterministic) {
      if(Ps::pVERBOSE || VERBOSE_CAR)
	cout << "stochastic choice of next direction " << endl;
      return choose_randomly(dirs.at(0), dirs.at(1));  
    }
    else {
      return dirs.at(1); // special for scenario 6: go north if east isn't green
    }
  }
}

Intersection::Turn Car::navigate() {
  if(Ps::pVERBOSE || VERBOSE_CAR)
    cout << "in navigate()" << endl;
  // plan route one step from current intersec to dest
  // TODO: check this can be done from constructor (uninitialized vars!)

  std::vector<Intersection::Direction> possible_dirs;
  plan_path_one_step(current_intersection, destination, possible_dirs);
  outgoing_direction = decide_next_outgoing_dir(possible_dirs); 
  if(Ps::pVERBOSE || VERBOSE_CAR)
    cout << "outgoing_direction " << outgoing_direction << endl;
  // Note: this could be left out when navigate() is called from constructor
  Intersection::Turn curr_turn = calculate_turn(current_incoming_direction,
						outgoing_direction);
  next_intersection = current_intersection->get_neighbouring_intersection
    (outgoing_direction);
  if(Ps::pVERBOSE || VERBOSE_CAR) {
    cout << "next_intersection: "; 
    next_intersection->print_id();
    cout << ", current turn: " << curr_turn << endl;
  }
  next_incoming_direction = (Intersection::Direction) 
    ((outgoing_direction + 2) % DIRECTIONS);
  if(Ps::pVERBOSE || VERBOSE_CAR)
    print_info();
  return curr_turn;
}
