// These options are transferred to the class Ps.cc, allowing run-time modification of them.
// To add a run-time option to class Ps.cc, just add another #define here.
// Comments after the definition will become the usage help information for that option.
// Do not put a #define here unless it makes sense to change it in run time.
// Put constant #defines in definitions.hh

#define ROAD_LENGTH 5 // Number of time steps a car needs to travel down this road
#define ROAD_CAPACITY 20 // Maximum cars allowed per road segment 
#define MAX_PHASE_DURATION 13 // Maximum steps a phase can be green
#define MAX_CYCLE_LENGTH 16 // Max. cycle length for SCATS and other ordered controllers
#define RPC_CYCLE_LENGTH 16 // (Fixed) cycle length for random phase controllers like the new NN
#define VERBOSE false // Debug output for the traffic simulator
#define GUI false // Should we launch the GUI
#define SEED 0 // Random seed for this run
#define BEST_PARAMS_FILE "best.params" // File to save best parameters in
#define PARAM_FILE "regular.saves.params" // File for regular saves of params
#define RESTORE_FILE "regular.saves.params" // File to read parameters from 
#define RESTORE false // Load parameters in RESTORE_FILE 
#define LEARN true // Whether we should learn or just run the current policy
#define BIAS 0.0 // biasing the NN towards an equal distribution of phase lengths

#define DISCOUNT 0.98 // Policy gradient discount factor
#define STEP_SIZE 0.0000001 // Policy gradient step size
#define STEPS_PER_EPOCH 100000 // Steps over which we report the long-term average
#define MAX_STEPS 0  // Maximum steps for optimisation. 0=unlimited
#define MAX_TIME 0 // Maximum time for optimisation. 0=unlimited
#define HIDDEN 0 // Number of hidden units for MLPs. If 0, use linear network
#define MAX_RAND 0.0 // Init all params randomly to at most this value.

#define USE_CURRENT_PHASE_INFO true // Whether the info what phase an intersection is in currently is included into the observations for the NN
#define USE_CYCLE_DURATION true // Whether NN knows time step in RPC cycle (0 - RPC_CYCLE_LENGTH)
#define USE_CURRENT_PHASE_DURATION_INFO true // Whether NN knows how long the current phase has been on 
#define USE_PHASE_DURATION_INFO true // Whether NN knows how long every phase has been on in this cycle
#define USE_DETEC_ACTIVE_INFO true // Whether NN knows which detectors currently have cars on them
#define USE_DETEC_HISTORY_INFO true // Whether NN knows all detector counts during this cycle
#define USE_NEIGHBOUR_INFO true // Whether NN uses detector information from neighbouring intersections
#define USE_PHASE_HISTORY_INFO false // Whether NN knows about the past policy of this intersection (how much time was spent on average in each phase)

#define CONTROLLER 0 // 0=neural net, 1=fixed, 2=scats, 4=mixed, 5=random next phase NN
#define ROWS 5 // Number of rows in traffic grid
#define COLS 5 // Number of cols in traffic grid
#define SCENARIO 0 // Scheme that dictates production of cars each time step (origin/destination, and possibly production rate)
#define CARS_PER_TIME_STEP 8 // For some scenarios, sets production rate
#define SUDDEN_INFLUX 10 // Parameter for sudden influx scenario

#define FIXED_PHASE_LENGTH 2 // Phase length for  fixed phase controller, also used as initial values for SCATS controller
#define UPPER_THRESHOLD 0.35 // Parameter for SCATS controller
#define LOWER_THRESHOLD 0.15 // Parameter for SCATS controller

//#define NN_TARGET 1.0 // Training amount for NN pre-learning
#define PRE_LEARNING_STEPSIZE 0.0001 
#define PRE_LEARNING_TIME 1000 // Iterations for SCATS-training of the NN

#define BOOST_SCATS false // Whether SCATS is allowed to make sudden changes to its policy (which would help in most scenarios)
#define LOCAL_REWARDS true
#define QUEUE_LENGTH_REWARDS false // LOCAL_REWARDS has to be true for this to have any effect, then it determines whether throughput or queue lengths are used as local rewards
#define BASELINE true // Whether a reward baseline is used in the learning algorithm
#define OFFSET 0 // In what time step do we start producing cars
#define SAVE_INTERSECTION_POLICY 12 // Save policy for this intersection
#define SAVE_FROM -1 // Save policy from this time step on (for 800 time steps). If -1, default numbers for SCATS and NN are used. If -2, no saving is done.
#define GUI_SPEED 900000 // microseconds for GUI sleep

// Natural actor critic stuff
#define NAC false // Use Natural Actor Critic instead of OLPOMDP
#define NAC_DISCOUNT 0.9 // Natural Actor Critic needs discount for critic value estimates

#define SCATS_CYCLE_EXTENSION false // some SCATS param to play with...
#define SCATS_GOAL_SATURATION 0.9 // saturation aimed for by SCATS
#define GREEDY_DRIVERS true // whether drivers take first possible green
#define GUI_DELAY 0 // time after which GUI comes on if GUI is true


// Line search stuff
#define LINE_SEARCH_STEPS 5000 // Number of steps for each grad estimate during line search
#define TOLERANCE 1e-9 // Convergence test. Keep going till grad is less than this
#define BACK_OFF 3.0 // how far to back off the step size when line search ends
#define EXPONENT_BASE 2.0 // Factor for step size increment in line search
#define DOWNHILL_THRESH 0.05 // Maximum  gone downhill before reverting parameters
#define MAX_TRIES 16 // Maximum number of line search iterations

// For CRF special
#define STEP_EDGE -1.0 // Stepsize for edges
#define STEP_ENODE -1.0 // Stepsize for extra node features
#define MAX_TREE_NO -10 // Number of trees to be sampled
#define MAX_TREE_TIME -10 // Maximum time allowable for tree sampling
#define MAX_RAND_EDGE -0.1 // Maximum edge feature random value
#define MAX_DELAY 3
#define MIN_DELAY 1


#define EDGE_REWARD_DEF 1 // Edge reward definition (how to get edge reward from two end nodes' reward
#define EDGE_FEA_SEL 1 // Edge feature selector
#define TIME_MOD 4 // Mod of time for offset scenario
#define EXIT_CHECK_WINDOW 500
#define THROUGHPUT_REWARDS 1
#define EXIT_TOLERANCE 0.001
