#include <math.h>

#include "PGBasics.hh"

#include "definitions.hh"
#include "Ps.hh"
#include "intersection.hh"
#include "car.hh"
#include "sts.hh"
#include "scenario.hh"

using namespace libpg;

Scenario::Scenario(int rows, int cols) : grid_rows(rows), grid_cols(cols) {
  //intersections = make_intersection_grid(grid_rows, grid_cols);
  //no_intersections = grid_rows * grid_cols;
}

void Scenario::produce_cars(STS* sts) {

  switch(Ps::pSCENARIO) {

  case Scenario::EW_BIASED: // 0
    produce_EW_biased_cars
      (sts, Ps::pCARS_PER_TIME_STEP, false); 
    break;

  case Scenario::RANDOM: // 1
    produce_random_cars(sts, Ps::pCARS_PER_TIME_STEP);
    break;
    
  case Scenario::SUDDEN_INFLUX: // 2
    produce_sudden_influx_cars(sts, Ps::pCARS_PER_TIME_STEP);
    break;

  case Scenario::EW_BIASED_CENTER_ONLY: // 3 
    produce_EW_biased_cars(sts, Ps::pCARS_PER_TIME_STEP, true);
    break;

  case Scenario::ONE_ROW_OFFSET_TEST: // 4
    one_row_offset_test(sts, Ps::pCARS_PER_TIME_STEP);
    break;
      
  case Scenario::SINUSOIDAL: // 5
    sinusoidal(sts, Ps::pCARS_PER_TIME_STEP);
    break;

  case Scenario::MAIN_STREET: // 6
    main_street(sts, Ps::pCARS_PER_TIME_STEP);
    break;

  case Scenario::BIG_GRID:
    big_grid(sts);
    break;

  default:
    cout<<"!! Unknown scenario type "<<Ps::pSCENARIO;
    exit(EXIT_FAILURE);
    
  }
}

void Scenario::big_grid_initialize() {
  // assumes full grid!
  int no_intersecs = grid_rows * grid_cols;
  double r;
  for(int i = 0; i < no_intersecs; i++) {
    int d = i;
    while(d == i) {
      d = (int) (random() / (double) RAND_MAX * no_intersecs);
    }
    big_grid_first_dest.push_back(d);
    d = i;
    while(d == i) {
      d = (int) (random() / (double) RAND_MAX * no_intersecs);
    }
    big_grid_second_dest.push_back(d);

    r = (random() / (double) RAND_MAX / 4);
    big_grid_first_prob.push_back(r);
    r = (random() / (double) RAND_MAX / 4);
    big_grid_second_prob.push_back(r);
  }
  
  double avg_cars = 0;
  //cout << "car destinations per intersection and production probs: " << endl;
  for(int i = 0; i < no_intersecs; i++) {
    //cout << i << " " << big_grid_first_dest.at(i) << ": " << big_grid_first_prob.at(i) << ", " << 
    //big_grid_second_dest.at(i) << ": " << big_grid_second_prob.at(i) << endl;
    avg_cars += big_grid_first_prob.at(i) + big_grid_second_prob.at(i);
  }
  cout << "Avg number of cars produced per time step: " << avg_cars << endl;
}


void Scenario::big_grid(STS* sts) {
  int no_intersecs = grid_rows * grid_cols;
  double r;
  std::vector<int>::iterator fd = big_grid_first_dest.begin();
  std::vector<int>::iterator sd = big_grid_second_dest.begin();
  std::vector<double>::iterator fp = big_grid_first_prob.begin();
  std::vector<double>::iterator sp = big_grid_second_prob.begin();

  for(int i = 0; i < no_intersecs; i++) {
    r = random() / (double) RAND_MAX ;
    if(r < *fp) {
      sts->produce_car(i, *fd);
    }
    r = random() / (double) RAND_MAX ;
    if(r < *sp) {
      sts->produce_car(i, *sd);
    }
    fd++; fp++; sd++; sp++;
  }
}

double rad2deg(int r) {
#define pi 3.14159265358979323846
  return r * 180.0 / pi; 
}

void Scenario::main_street(STS* sts, int num) {  

  /*                           20(0)
                                    |
  01(1) --  11(2) -- 21(3) -- 31(4)
                      |           |
                 12(5) -- 22(6)
                      |           |
                 13(7)    23(8) */

  int top_main_road = 0;
  int left_main_road = 1;
  int right_main_road = 4;
  int bottom_main_road = 8;
  int soccer_stadium = 7;

  //int offset = Ps::pOFFSET;
  //int time_step = ((int) sts->sim_time) + offset;

  sts->produce_car(top_main_road, bottom_main_road);
  sts->produce_car(bottom_main_road, top_main_road);
  sts->produce_car(left_main_road, right_main_road);
  sts->produce_car(right_main_road, left_main_road);

  sts->produce_car(soccer_stadium, right_main_road);

  
  double r = random() / (double) RAND_MAX;
  //double prob = 0.15;
  double prob = 0.1;
  if(r  < prob) {
    //sts->produce_car(top_main_road, bottom_main_road);
    sts->produce_car(bottom_main_road, top_main_road); 
  }
  r = random() / (double) RAND_MAX;
  if(r  < prob) {
  sts->produce_car(left_main_road, right_main_road); 
  //sts->produce_car(right_main_road, left_main_road);
  }
  r = random() / (double) RAND_MAX;
  if(r  < prob) {
    sts->produce_car(soccer_stadium, right_main_road); 
  }
 
  /*
  r = random() / (double) RAND_MAX;
  if(r < 0.1) {
    sts->produce_car(top_main_road, right_main_road);
  }
  */
  /*
  if(r  < 0.33) 
    sts->produce_car(left_main_road, top_main_road);
  else if(r >= 0.33 && r < 0.66)
    sts->produce_car(left_main_road, right_main_road);
  else if(r >= 0.66)
    sts->produce_car(left_main_road, bottom_main_road);
*/
  /*
  sts->produce_car(top_main_road, bottom_main_road);
  sts->produce_car(bottom_main_road, top_main_road);
  sts->produce_car(left_main_road, right_main_road);
  sts->produce_car(right_main_road, left_main_road);
  sts->produce_car(soccer_stadium, right_main_road);
  if(time_step % 5 == 0) { 
    sts->produce_car(top_main_road, bottom_main_road);
    sts->produce_car(bottom_main_road, top_main_road);
    sts->produce_car(left_main_road, right_main_road);
    sts->produce_car(right_main_road, left_main_road);
    sts->produce_car(soccer_stadium, right_main_road);
 }
  */
  //sts->produce_car(soccer_stadium, left_main_road);
  //sts->produce_car(soccer_stadium, left_main_road);
}

void Scenario::sinusoidal(STS* sts, int num) {  
#define pi 3.14159265358979323846

  int offset = Ps::pOFFSET;
  int time_step = ((int) sts->sim_time) + offset;

  int num1 = 3;
  //if(Ps::pVERBOSE) {
  //cout << "time step: " << time_step << ", ";
    //}
  //double seed = rad2deg(time_step);
  double seed = 2 * pi  / 20 * time_step;
  //cout << "seed: " << seed << endl;

  double sin1 = (sin(seed) + 1) * num1 / 2;
  double cosin1 = (cos(seed) + 1) * num1 / 2;

  //cout << "sin: " << sin1 << ", cosin:" <<  cosin1 << endl;
  sin1 = (int) (sin1 + 0.5);
  //cosin1 = (int) (cosin1 + 0.5);
  cosin1 = num1 - sin1;

  //if(Ps::pVERBOSE) {
  //cout << "cars: " << sin1  << ", " 
  //<<  cosin1 << ",   ";
    //}

  int o1 = grid_rows / 2; 
  int d1 = grid_rows / 2 + grid_cols - 1;
  int o2  = 0; 
  int d2 = grid_rows + grid_cols - 2;

  for(int i = 0; i < sin1; i++) {
    sts->produce_car(o1, d1);
    //sts->produce_car(d1, o1);
  }
  for(int i = 0; i < cosin1; i++) {
    sts->produce_car(o2, d2); 
    //sts->produce_car(d2, o2); 
  } 
}

void Scenario::produce_sudden_influx_cars(STS* sts, int num) {
  //int offset = Ps::pOFFSET;
  //int time_step = ((int) sts->sim_time) + offset;

  int o1 = grid_rows / 2; 
  int d1 = grid_rows / 2 + grid_cols - 1;
  int o2  = 0; 
  int d2 = grid_rows + grid_cols - 2;

  /*
  int o1  = 10; 
  int d1 = 14; 
  int o2 = 2; 
  int d2 = 22;
  */
  /*
  int o1  = 6;
  int d1 = 8;
  int o2 = 1;
  int d2 = 13;
  */
  for(int i = 0; i < num; i++) {
    sts->produce_car(o1, d1);
    //sts->produce_car(d1, o1);
  }

  //if(time_step % 2 == 0) { 
  // sts->produce_car(o2, d2); 
  // sts->produce_car(d2, o2); 
  //} 
 
  double r = random() / (double) RAND_MAX;
  //if(time_step % 20 == 0 && time_step > 0) {// sudden influx 
  if(r  < 0.02) {
    for(int i = 0; i < Ps::pSUDDEN_INFLUX; i++) {  
      sts->produce_car(o2, d2); 
    }
  }
 
}  

void Scenario::one_row_offset_test(STS* sts, int num) {
  // assumes a 1 x n grid
  int offset = Ps::pOFFSET;
  int time_step = ((int) sts->sim_time) + offset;
  int o1  = 0;
  int d1 = grid_cols - 1;
 
  if(!Ps::pTIME_MOD || time_step % Ps::pTIME_MOD == 0) {
    for(int i = 0; i < num; i++) {
      sts->produce_car(o1, d1);
    }
    }
}  

void Scenario::produce_random_cars(STS* sts, int num) {
  // assumes full grid!
  int no_intersections = grid_rows * grid_cols;
  //srand(time(NULL));
  for(int i = 0; i < num; i++) {
    int o = random2() % no_intersections;
    int d = random2() % no_intersections;
    while(o == d)
      d = random2() % no_intersections;
    sts->produce_car(o, d);
  }
}

void Scenario::produce_EW_biased_cars(STS* sts, int num, bool center_only) {
  /* Produces num cars that travel east or west with prob. 0.75 (and south or
     north with prob. 0.25). Cars are produced on the edges of the grid and travel
     to the opposite edge; where on the edge they are produced is random
     if center_only is false, else they travel through a center intersection. */

 for(int i = 0; i < num; i++) {
   int edge, from_row, from_col, to_row, to_col;
   double r = random2() / (double) RAND_MAX;
   if(r > 0.75 && grid_rows > 1 ||grid_cols <= 1 ) {
     edge = 0; // North
   }
   else if(grid_cols > 1) {
     edge = 1; // East
   }
   else {
     cout << "Scenario::produce_EW_biased_cars(): cannot produce car" << 
                     " due to grid layout!" << endl;
     return;
   }
   r = random2() / (double) RAND_MAX;
   if(r > 0.5) {
     edge += 2;
   }
   if(edge == 0 || edge == 2) { // travelling N-S or S-N
     if(!center_only)
       from_col = random2() % grid_cols;
     else
       from_col = grid_cols / 2;
     to_col = from_col;
     if(edge == 0) {
       from_row = 0;
       to_row = grid_rows - 1;
     }
     else {
       from_row = grid_rows - 1;
       to_row = 0;
     }
   }
   else { // travelling E-W or W-E
     if(!center_only)
       from_row = random2() % grid_rows;
     else
       from_row = grid_rows / 2;
     to_row = from_row;
     if(edge == 1) {
       from_col = grid_cols - 1;
       to_col = 0;
     }
     else {
       from_col = 0;
       to_col = grid_cols - 1;
     }
   } 
   int origin = from_row * grid_cols + from_col;
   int destination = to_row * grid_cols + to_col;   
   sts->produce_car(origin, destination);
 }
}

//void STS::produce_car(int origin, int destination) {
 
void Scenario::make_intersections(IntersectionVec& intersections,
				      int rows, int cols) {
  switch(Ps::pSCENARIO) {

  case Scenario::ONE_ROW_OFFSET_TEST: // 4
    make_intersection_row(intersections, rows, cols);
    break;
          
  case Scenario::SUDDEN_INFLUX: 
    make_intersection_cross(intersections, rows, cols);
    break;

  case Scenario::EW_BIASED_CENTER_ONLY:
    make_intersection_cross(intersections, rows, cols);
    break;
      
  case Scenario::SINUSOIDAL:
    make_intersection_cross(intersections, rows, cols);
    break;

  case Scenario::MAIN_STREET:
    make_main_street_intersections(intersections);
    break;

  case Scenario::BIG_GRID:
    make_intersection_grid(intersections, rows, cols);
    big_grid_initialize();
    break;

  default:
   make_intersection_grid(intersections, rows, cols);
  }
}

void Scenario::make_intersection_row(IntersectionVec& intersections, int rows, int cols) {
  if(rows != 1) {
    cout << "Rows should be one for this scenario!" << endl;
  }
  make_intersection_grid(intersections, 1, cols);
  intersections.at(0)->dummy = true;
  intersections.at(cols - 1)->dummy = true;
}

void Scenario::make_main_street_intersections(IntersectionVec& intersections) {
  /*                           20(0)
                                    |
  01(1) --  11(2) -- 21(3) -- 31(4)
                      |           |
                 12(5) -- 22(6)
                      |           |
                 13(7)    23(8) */
  Intersection* i20 = new Intersection(2,0);   intersections.push_back(i20);  i20->dummy = true;
  Intersection* i01 = new Intersection(0,1);   intersections.push_back(i01);  i01->dummy = true;
  Intersection* i11 = new Intersection(1,1);   intersections.push_back(i11);  
  Intersection* i21 = new Intersection(2,1);   intersections.push_back(i21);
  Intersection* i31 = new Intersection(3,1);   intersections.push_back(i31);  i31->dummy = true;
  Intersection* i12 = new Intersection(1,2);   intersections.push_back(i12);  
  Intersection* i22 = new Intersection(2,2);   intersections.push_back(i22);  
  Intersection* i13 = new Intersection(1,3);   intersections.push_back(i13);  i13->dummy = true;
  Intersection* i23 = new Intersection(2,3);   intersections.push_back(i23);  i23->dummy = true;
  
  i20->incoming[Intersection::South] = i21;
  i20->outgoing[Intersection::South] = i21;

  i01->incoming[Intersection::East] = i11;
  i01->outgoing[Intersection::East] = i11;

  i11->incoming[Intersection::East] = i21; 
  i11->outgoing[Intersection::East] = i21;
  i11->incoming[Intersection::South] = i12; 
  i11->outgoing[Intersection::South] = i12;
  i11->incoming[Intersection::West] = i01;
  i11->outgoing[Intersection::West] = i01;

  i21->incoming[Intersection::North] = i20;
  i21->outgoing[Intersection::North] = i20;
  i21->incoming[Intersection::East] = i31;
  i21->outgoing[Intersection::East] = i31;
  i21->incoming[Intersection::South] = i22;
  i21->outgoing[Intersection::South] = i22;
  i21->incoming[Intersection::West] = i11;
  i21->outgoing[Intersection::West] = i11;

  i31->incoming[Intersection::West] = i21;
  i31->outgoing[Intersection::West] = i21;

  i12->incoming[Intersection::North] = i11;
  i12->outgoing[Intersection::North] = i11;
  i12->incoming[Intersection::East] = i22;
  i12->outgoing[Intersection::East] = i22;
  i12->incoming[Intersection::South] = i13;
  i12->outgoing[Intersection::South] = i13;

  i22->incoming[Intersection::North] = i21;
  i22->outgoing[Intersection::North] = i21;
  i22->incoming[Intersection::West] = i12;
  i22->outgoing[Intersection::West] = i12;
  i22->incoming[Intersection::South] = i23;
  i22->outgoing[Intersection::South] = i23;

  i13->incoming[Intersection::North] = i12;
  i13->outgoing[Intersection::North] = i12;

  i23->incoming[Intersection::North] = i22;
  i23->outgoing[Intersection::North] = i22;

  IntersectionVec::iterator i;
  for(i = intersections.begin(); i < intersections.end(); i++) {
    for(int d = 0; d < DIRECTIONS; d++) {
      (*i)->length_incoming[d] =  Ps::pROAD_LENGTH;
      (*i)->capacity_incoming[d] = Ps::pROAD_CAPACITY;
    }
  }

}

void Scenario::make_intersection_grid(IntersectionVec& intersections,
				      int rows, int cols) {
  Intersection* ints_grid[rows][cols];
  for(int i = 0; i < rows; i++) 
    for(int j = 0; j < cols; j++) 
      ints_grid[i][j] = new Intersection(j,i);
  
  for(int i = 0; i < rows - 1; i++) 
    for(int j = 0; j < cols; j++) {
      ints_grid[i][j]->incoming[Intersection::South] = ints_grid[i + 1][j];
      ints_grid[i][j]->length_incoming[Intersection::South] = 
	Ps::pROAD_LENGTH;
      ints_grid[i][j]->capacity_incoming[Intersection::South] = 
	Ps::pROAD_CAPACITY;
      ints_grid[i][j]->outgoing[Intersection::South] = ints_grid[i + 1][j];
    }

  for(int i = 0; i < rows; i++) 
    for(int j = 1; j < cols; j++) {
      ints_grid[i][j]->incoming[Intersection::West] = ints_grid[i][j - 1];
      ints_grid[i][j]->length_incoming[Intersection::West] = 
	Ps::pROAD_LENGTH;
      ints_grid[i][j]->capacity_incoming[Intersection::West] = 
	Ps::pROAD_CAPACITY;
      ints_grid[i][j]->outgoing[Intersection::West] = ints_grid[i][j - 1];
    }

  for(int i = 1; i < rows; i++) 
    for(int j = 0; j < cols; j++) {
      ints_grid[i][j]->incoming[Intersection::North] = ints_grid[i - 1][j];
      ints_grid[i][j]->length_incoming[Intersection::North] = 
	Ps::pROAD_LENGTH;
      ints_grid[i][j]->capacity_incoming[Intersection::North] = 
	Ps::pROAD_CAPACITY;
      ints_grid[i][j]->outgoing[Intersection::North] = ints_grid[i - 1][j];
    }

  for(int i = 0; i < rows; i++) 
    for(int j = 0; j < cols - 1; j++) {
      ints_grid[i][j]->incoming[Intersection::East] = ints_grid[i][j + 1];
      ints_grid[i][j]->length_incoming[Intersection::East] = 
	Ps::pROAD_LENGTH;
      ints_grid[i][j]->capacity_incoming[Intersection::East] = 
	Ps::pROAD_CAPACITY;
      ints_grid[i][j]->outgoing[Intersection::East] = ints_grid[i][j + 1];
    }
  //IntersectionVec intersections;
  for(int i = 0; i < rows; i++) 
    for(int j = 0; j < cols; j++) 
      intersections.push_back(ints_grid[i][j]);
  //return intersections;
}

void Scenario::make_intersection_cross(IntersectionVec& intersections,
				      int rows, int cols) {

  //cout << "rows, cols: "<< rows << " " << grid_rows << " " << cols << " " << grid_cols << endl;
  Intersection* ints_grid[rows][cols];
  for(int i = 0; i < rows; i++) 
    for(int j = 0; j < cols; j++) 
      ints_grid[i][j] = new Intersection(j,i);

  for(int i = 0; i < rows - 1; i++) 
    for(int j = 0; j < cols; j++) {
      ints_grid[i][j]->incoming[Intersection::South] = ints_grid[i + 1][j];
      ints_grid[i][j]->length_incoming[Intersection::South] = 
	Ps::pROAD_LENGTH;
      ints_grid[i][j]->capacity_incoming[Intersection::South] = 
	Ps::pROAD_CAPACITY;
      ints_grid[i][j]->outgoing[Intersection::South] = ints_grid[i + 1][j];
    }

  for(int i = 0; i < rows; i++) 
    for(int j = 1; j < cols; j++) {
      ints_grid[i][j]->incoming[Intersection::West] = ints_grid[i][j - 1];
      ints_grid[i][j]->length_incoming[Intersection::West] = 
	Ps::pROAD_LENGTH;
      ints_grid[i][j]->capacity_incoming[Intersection::West] = 
	Ps::pROAD_CAPACITY;
      ints_grid[i][j]->outgoing[Intersection::West] = ints_grid[i][j - 1];
    }

  for(int i = 1; i < rows; i++) 
    for(int j = 0; j < cols; j++) {
      ints_grid[i][j]->incoming[Intersection::North] = ints_grid[i - 1][j];
      ints_grid[i][j]->length_incoming[Intersection::North] = 
	Ps::pROAD_LENGTH;
      ints_grid[i][j]->capacity_incoming[Intersection::North] = 
	Ps::pROAD_CAPACITY;
      ints_grid[i][j]->outgoing[Intersection::North] = ints_grid[i - 1][j];
    }

  for(int i = 0; i < rows; i++) 
    for(int j = 0; j < cols - 1; j++) {
      ints_grid[i][j]->incoming[Intersection::East] = ints_grid[i][j + 1];
      ints_grid[i][j]->length_incoming[Intersection::East] = 
	Ps::pROAD_LENGTH;
      ints_grid[i][j]->capacity_incoming[Intersection::East] = 
	Ps::pROAD_CAPACITY;
      ints_grid[i][j]->outgoing[Intersection::East] = ints_grid[i][j + 1];
    }

  //IntersectionVec intersections;
  for(int i = 0; i < rows; i++) 
    for(int j = 0; j < cols; j++) 
      
      if(i == rows / 2 || j ==  cols / 2) {
	intersections.push_back(ints_grid[i][j]);
	if(i == 0 || j == 0 || i == rows - 1 || j == cols - 1) {
	  ints_grid[i][j]->dummy = true;
	}
	//cout << "producing intersection " << j << " " << i << endl;
      }
      else {
	//cout << "trying to delete: " <<  j << " " << i << endl;
	delete ints_grid[i][j];
      }
      
  //return intersections;
}
