typedef list<Car*> CarList;
typedef std::vector<Intersection*> IntersectionVec;

class Scenario;

class STS: public libpg::Simulator {

private:
  void get_observation_for_learning_controller(libpg::Observation& obs); 
  void get_observation_for_scats_controller(libpg::Observation& obs); 
  void get_observation_for_mixed_controller(libpg::Observation& obs); 
  

    int obsRows;
    int obsCols;
    int agents;
    int actionDim;
    int actions;
  int no_controlled_intersections; // non-dummy intersections;
  std::vector<string> statisticsNames;
  Vector statistics;

  int save_history_from;
  int save_history_to;
  string intersec_summary_file_name;
  int save_history_for_intersection;
  bool save_detailed_history;
  ofstream* history_file;
  ofstream* stat_file;
  
  IntersectionVec tracked_intersections;
  int * current_phase;

public:
  // constants should not have same name as classes!
    enum STSControllerEnum {_LearningController, _FixedPhaseController, 
			    _ScatsController, _FixedController, 
			    _MixedController, _RandomPhaseController, 
			    _Scenario6Controller, _Scenario4Controller,
			    _MixedRPController, _CRFController, _FixedRPController};

  STS(Scenario* sc);
  //STS(CarList the_cars, IntersectionVec the_intersections);
  //STS(int rows, int cols);
  virtual ~STS() {}

  Scenario* scenario;
  CarList cars;
  IntersectionVec intersections;
  int grid_rows;
  int grid_cols;
  double sim_time;
  double next_id; // keep track of car ids in the system
  int cars_arrived_last_time_step;
  int cars_failed_last_time_step; // cars that could not be entered into system
  double sum_traveltime_last_time_step;
  double reward_last_time_step; // have reset

  // Abstract methods inherited from simulator that must be made concrete.
  double get_global_reward();
  void getReward(Vector& reward);
  void getObservation(libpg::Observation& obs); 
  int doAction(Vector& action);
  int getMaxEpisodeLength();
  int getObsRows();
  int getObsCols();
  int getAgents();
  int getActionDim();
  int getRewardDim();

  void reset(); // delete cars and reset sim_time...
  std::vector<string>& STS::getStatisticsNames();
  Vector& STS::getStatistics(int steps);

  void update_intersections(Vector& phase_changes);
  void update_cars();
  void update_reward(Car* car); 
  void save_statistics();
  void clearStatistics(); 
  void track_intersections(IntersectionVec& track){tracked_intersections = track;}  
  void save_history(int from, int to, string file);
  void write_snapshot();
  void write_tracked_intersections_statistics();
  //void produce_random_cars(int num);
  //void produce_EW_biased_cars(int num);
  void produce_car(int origin, int destination);
  void print_map();
  void run();
  IntersectionVec& get_intersections() {return intersections;}
  Intersection* get_intersection(int i) {return intersections.at(i);}
  CarList& get_cars() {return cars;}
  int get_rows() {return grid_rows;}
  int get_cols() {return grid_cols;}
    
  //Scenario::IntersectionVec make_intersection_grid(int rows, int cols);
  void common_init();

  virtual int * get_last_action();
  virtual char *get_diagnosis();
};
