#include "PGBasics.hh"
#include "TreeBasics.hh"
#include "test_traffic.hh"

#include "definitions.hh"
#include "Ps.hh"
#include "intersection.hh"
#include "car.hh"
#include "sts.hh"
#include "OLPomdp.hh"
#include "GPomdp.hh"
#include "LineSearchAlg.hh"
#include "STSOptim.hh"

#ifdef HAVE_QT
#include <qthread.h> // needed for testgui.hh
#include <qapplication.h>
#include <qwidget.h> // needed for simgui.hh
#include <qthread.h>
#include "simgui.hh"
#include "testgui.hh"
#endif

using namespace libpg;


int test_traffic(int argc, char** argv)
{
   // Allow command line options
	Ps::processOpts(argc, argv);

	// Dump options to standard output
	Ps::print(cout);
	char fname[100];

	sprintf(fname, "%s.%d", "settings", getpid());

	ofstream* o = new ofstream(fname);
	Ps::print(*o);
	(*o) << "Process ID: " << getpid() << endl;
	o->close();

	STSOptim stsOptim(argc, argv);

	if (!Ps::pGUI) 
	{
		if (Ps::pLEARN) 
			stsOptim.learn(Ps::pSTEPS_PER_EPOCH, 
							Ps::pMAX_TIME, 
							Ps::pMAX_STEPS
							);
		else 
			stsOptim.evaluateAndPrint(Ps::pSTEPS_PER_EPOCH,
					   Ps::pMAX_STEPS);  
	}
	else 
	{
		#ifdef HAVE_QT 
		cout << "I will display a GUI... " << endl;

		if(Ps::pGUI_DELAY > 0) 
		{
			if (Ps::pLEARN) 
				stsOptim.learn(Ps::pGUI_DELAY, 
							   Ps::pMAX_TIME, 
							   Ps::pGUI_DELAY
							   );
			else 
				stsOptim.evaluateAndPrint(Ps::pGUI_DELAY, 
				    Ps::pGUI_DELAY);				      
		}

		QApplication app(argc, argv); 

		STS& sim = *(stsOptim.get_simulator());
		WidgetTest test(sim); 
		app.setMainWidget(&test); 
		test.show();

		//TestGui tg(stsOptim);
		TestGui tg(stsOptim, &test);
		tg.start();
		app.exec();

		#else
			cout<<"Can't do this without QT!\n";
		#endif
	}
	
	exit(EXIT_SUCCESS);	
}
