#include "TreeBasics.hh"

using namespace std;

namespace tree_mcmc {

CRFCircle::CRFCircle(int *argc, char** argv):CRFApprox(argc, argv)
{

}

void CRFCircle::init_graph(int *argc, char **argv)
{
	Utility::get_arg(&global_num_node, argc, argv, "-na", 1, 0);

	printf("In CRF for circle, Num Agent = %d\tNum Class = %d\n", global_num_node, num_class);

	global_num_edge = global_num_node;
	num_row = 2;
	num_col = global_num_node / 2;

	alloc_adj_matrix();
	alloc_potentials();
	associate_edge_with_nodes();
	define_partitions();
}


void CRFCircle::alloc_adj_matrix()
{
	int i, j;

	adj_matrix = new int * [global_num_node + 1];

	for (i = 1; i <= global_num_node; i ++)
	{
		adj_matrix[i] = new int [global_num_node + 1];
		for (j = 1; j <= global_num_node; j ++)
			adj_matrix[i][j] = 0;
	}

	adj_matrix[1][2] = 1;
	for (i = 2; i < num_col; i ++)
		adj_matrix[i][i - 1] = adj_matrix[i][i + 1] = 1;
	adj_matrix[num_col][num_col - 1] = 1;

	adj_matrix[num_col + 1][num_col + 2] = 1;
	for (i = num_col + 2; i < global_num_node; i ++)
		adj_matrix[i][i - 1] = adj_matrix[i][i + 1] = 1;
	adj_matrix[global_num_node][global_num_node - 1] = 1;

	adj_matrix[num_col + 1][1] = adj_matrix[1][num_col + 1] = 1;
	adj_matrix[num_col][global_num_node] = adj_matrix[global_num_node][num_col] = 1;
}


void CRFCircle::associate_edge_with_nodes()
{
	int cur_item;
	Edge *p_edge;

	p_edge = edge_array + 1;
	for (cur_item = 1; cur_item < num_col; cur_item ++)
	{
		p_edge->p_node[1] = node_array + cur_item;
		p_edge->p_node[2] = node_array + cur_item + 1;
		p_edge ++;
	}

	for (cur_item ++; cur_item < global_num_node; cur_item ++)
	{
		p_edge->p_node[1] = node_array + cur_item;
		p_edge->p_node[2] = node_array + cur_item + 1;
		p_edge ++;
	}

	p_edge->p_node[1] = node_array + 1;
	p_edge->p_node[2] = node_array + num_col + 1;
	p_edge ++;
	p_edge->p_node[1] = node_array + num_col;
	p_edge->p_node[2] = node_array + global_num_node;
}


int CRFCircle::get_edge_id(int node1, int node2)
{
	if(node1 < node2)
	{
		if (node1 + 1 == node2)		// horizontal edge
		{
			if (node1 < num_col)
				return node1;
			else
				return node1 - 1;
		}
		else						// vertical edge
		{
			if (node1 == 1)
				return global_num_node - 1;
			else
				return global_num_node;
		}
	}
	else
	{
		if (node2 + 1 == node1)		// horizontal edge
		{
			if (node2 < num_col)
				return node2;
			else
				return node2 - 1;
		}
		else						// vertical edge
		{
			if (node2 == 1)
				return global_num_node - 1;
			else
				return global_num_node;
		}
	}
}

void CRFCircle::define_partitions()
{
	int cur_part, next_index, quater_floor, quater_ceil, temp;
	Tree *p_tree;
	Partition *p_part;

	global_num_partition = 2;
	partitions = new Partition[3];
	quater_floor = int((num_col + 0.5) / 2);
	quater_ceil = num_col - quater_floor;

	for (cur_part = 1;  cur_part <= 2; cur_part ++)		// for the 4 partitions
	{
		p_part = partitions + cur_part;
		p_part->num_trees = 2;
		p_part->trees = new Tree[3];

#ifdef RB_NODE_SINGLE
		if (cur_part == 1)
		{
			p_part->trees[1].cal_node_rb = 1;
			p_part->trees[2].cal_node_rb = 1;
		}
		else
		{
			p_part->trees[1].cal_node_rb = 0;
			p_part->trees[2].cal_node_rb = 0;
		}
#else
		p_part->trees[1].cal_node_rb = 1;
		p_part->trees[2].cal_node_rb = 1;
#endif
	
		if (cur_part == 1)
		{
			// 1st, upper tree (chain)
			p_tree = p_part->trees + 1;
			p_tree->num_node = num_col;
			p_tree->local_to_global = new int[p_tree->num_node + 1];

			for (next_index = 1; next_index <= num_col; next_index++)
				p_tree->local_to_global[next_index] = next_index;

			// 2nd, lower tree
			p_tree = p_part->trees + 2;
			p_tree->num_node = num_col;
			p_tree->local_to_global = new int[p_tree->num_node + 1];

			for (next_index = 1; next_index <= num_col; next_index++)
				p_tree->local_to_global[next_index] = next_index + num_col;
		}
		else if (cur_part == 2)
		{
			// 1st, left tree
			p_tree = p_part->trees + 1;
			p_tree->num_node = num_col;
			p_tree->local_to_global = new int[p_tree->num_node + 1];

			for (next_index = 1; next_index <= quater_floor; next_index ++)
				p_tree->local_to_global[next_index] = next_index;
			temp = num_col + 1;
			for (; next_index <= num_col; next_index ++, temp ++)
				p_tree->local_to_global[next_index] = temp;

			// 2nd, right tree
			p_tree = p_part->trees + 2;
			p_tree->num_node = num_col;
			p_tree->local_to_global = new int[p_tree->num_node + 1];

			temp = quater_floor + 1;
			for (next_index = 1; next_index <= quater_ceil; next_index ++, temp ++)
				p_tree->local_to_global[next_index] = temp;	
			temp = num_col + quater_ceil + 1;
			for (; next_index <= num_col; next_index ++, temp ++)
				p_tree->local_to_global[next_index] = temp;	
		}
	}
}


void CRFCircle::fill_edge_reward(double *reward)
{
	int cur_edge;
	double *cur_reward;
	Edge *pEdge;

	cur_reward = reward + global_num_node + 1;
	pEdge = edge_array + 1;

	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{
		switch (edge_reward_def)
		{
		case 1:		// arithmetic mean of two end nodes (default)
			*(cur_reward ++) = (reward[pEdge->p_node[1] - node_array] + reward[pEdge->p_node[2] - node_array]) / 2;
			break;
		case 2:		// geometric mean
			*(cur_reward ++) = sqrt(reward[pEdge->p_node[1] - node_array] * reward[pEdge->p_node[2] - node_array]);
			break;
		case 3:		// min of two end nodes
			*(cur_reward ++) = min(reward[pEdge->p_node[1] - node_array], reward[pEdge->p_node[2] - node_array]);
			break;
		case 4:		// max of two end nodes
			*(cur_reward ++) = max(reward[pEdge->p_node[1] - node_array], reward[pEdge->p_node[2] - node_array]);
			break;
		case 5:		// end node [2]
			*(cur_reward ++) = reward[pEdge->p_node[2] - node_array];
			break;
		case 6:		// end node [1]
			*(cur_reward ++) = reward[pEdge->p_node[1] - node_array];
			break;
		default:
			cout<<"Unforseen edge reward def: "<<edge_reward_def<<endl;
			exit(1);
		}

		pEdge ++;
	}
}

}
