#include "TreeBasics.hh"


using namespace std;

namespace tree_mcmc {

CRFExact::CRFExact(int *argc, char** argv):CRFBasic(argc, argv)
{

}



/** 
 *	For CRF which is originally a tree, then work on this tree
 */
void CRFExact::build_ori_tree(int root_id)
{
	int cur_node, i, max_num_class;
	Node * p_node;

	// First allocate and define tree
	init_tree = new Tree;
	init_tree->cal_node_rb = true;

	init_tree->num_node = global_num_node;
	init_tree->local_to_global = new int[global_num_node + 1];

	for (cur_node = 1; cur_node <= global_num_node; cur_node++)
		init_tree->local_to_global[cur_node] = cur_node;

	build_tree(init_tree, root_id);

	for (cur_node = 1; cur_node <= global_num_edge; cur_node ++)
		init_tree->cal_edge_rb[cur_node] = true;

	if(verbosity & 1)
	{
		printf("********     Chain     ********\n");
		display_tree(init_tree, 7);		
	}

	// Finally allocate the msg_buffer for all nodes after obtaining the max_num_son information

	p_node = node_array + 1;
	max_num_class = 0;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		if (p_node->num_class > max_num_class)
			max_num_class = p_node->num_class;
		p_node->msg_buffer = new double * [p_node->max_num_son + 1];
		for (i = 1; i <= p_node->max_num_son; i++)
			p_node->msg_buffer[i] = new double [p_node->num_class + 1];
		p_node ++;
	}
	

	sample_buf = new double [max_num_class + 1];
	rb_buf = new double * [max_num_class + 1];
	for (cur_node = 1; cur_node <= max_num_class; cur_node ++)
		rb_buf[cur_node] = new double [max_num_class + 1];

	for (cur_node = 1; cur_node <= init_tree->num_node; cur_node ++)
		node_array[cur_node].num_son = init_tree->num_son[cur_node];
}


/**
 *	Perform inference on the whole graph (tree) using RB
 */
void CRFExact::infer_graph()
{
	// First define the initial tree
	clear_RB();

	cal_cond_pot(init_tree);
	forward_msg_pass(init_tree);
	infer_root(init_tree);
	backward_msg_pass(init_tree);

	if(verbosity & 6)	
		display_RB();

	normalize_RB();

	if(verbosity & 6)
		display_RB();
}


/**
 *	Perform sampling on the whole graph (tree) using RB
 */
void CRFExact::sample_graph()
{
	cal_cond_pot(init_tree);
	forward_msg_pass(init_tree);
	sample_root(init_tree);
	backward_sample(init_tree);
/*
	cout<<1<<endl;
	dump_graph("input.dat", "result.dat");
	cin.get();
*/
}


// A generic sampler initializer
void CRFExact::init_tree_sampler(int *argc, char **argv)
{	
	CRFBasic::init_tree_sampler(argc, argv);
	build_ori_tree();
}

}
