#include "TreeBasics.hh"


using namespace std;

namespace tree_mcmc {


/** 
 *	For CRF which is a chain, then work on this tree
 */

CRFSensorChain::CRFSensorChain(int *argc, char** argv):CRFChain(argc, argv)
{

}


/**
 *	After calculating the RB estimations on the graph,
 *	Calculate the feature expectations.
 */
void CRFSensorChain::cal_fea_expectation()
{
	int cur_edge, cur_node;
	double *rb1, *fept1, **fept2;
	Node *p_node;
	Edge *p_edge;


	// Node features
	cal_node_fea_exp();

	// Node Extra features
	fept2 = node_extra_fea_exp + 1;
	p_node = node_array + 1;
	for (cur_node = 1; cur_node <= global_num_node; cur_node ++)
	{
		rb1 = p_node->rb_est + 1;
		*(*(fept2 ++) + 1) = *(rb1 + 1) + *(rb1 + 2) + 2 * (*(rb1 + 3) + *(rb1 + 4) + *(rb1 + 5));
		p_node ++;
	}

	// Edge features
	p_edge = edge_array + 1;
	fept2 = edge_fea_exp + 1;

	for (cur_edge = 1; cur_edge <= global_num_edge; cur_edge ++)
	{
		fept1 = * (fept2 ++) + 1;
		*(fept1 ++) = (p_edge->rb_est[3][5] + p_edge->rb_est[4][5] + p_edge->rb_est[6][2] + 
						p_edge->rb_est[6][4] + p_edge->rb_est[6][5]) * edge_fea_intensity;				
		p_edge ++;
	}
}


void CRFSensorChain::fill_trace()
{
	int cur_item, sample1, sample2, offset;
	double **pd2, *pd1, **pd2_2, *pd1_2;
	Node *p_node;
	Edge *p_edge;

	// Node features
	fill_node_trace();

	// Node Extra features
	p_node = node_array + 1;
	pd2 = node_extra_fea_trace + 1;
	pd2_2 = node_extra_fea_exp + 1;
	for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
	{
		pd1_2 = *(pd2_2 ++) + 1;

		*pd1_2 = - *pd1_2;
		if (p_node->sampled_value > 3)
			*pd1_2 += 2;
		else if (p_node->sampled_value > 1)
			*pd1_2 += 1;

		*(*(pd2 ++) + 1) += *pd1_2;

		p_node ++;
	}

	// Edge features
	pd2 = edge_fea_trace + 1;
	pd2_2 = edge_fea_exp + 1;
	p_edge = edge_array + 1;
	offset = 1;

	for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)
	{
		pd1 = *(pd2 ++) + offset;
		pd1_2 = *(pd2_2 ++) + offset;
		sample1 = p_edge->p_node[1]->sampled_value;
		sample2 = p_edge->p_node[2]->sampled_value;

		if (sample1 == 6 && (sample2 == 2 || sample2 == 4 || sample2 == 5)
			|| sample2 == 5 && (sample1 == 3 || sample1 == 4))
		{
			*pd1_2 = edge_fea_intensity - *pd1_2;				
			*(pd1++) += *(pd1_2 ++);
		}
		else
		{
			*pd1_2 = - *pd1_2;
			*(pd1++) += *(pd1_2 ++);
		}
		p_edge ++;
	}
}


/**
 *	Set the edge and node potentials of the graph.  
 *	Space allocation has been done by alloc_potentials()
 */
void CRFSensorChain::set_potentials()
{
	int cur_item;
	Node *p_node;
	Edge *p_edge;
	double *cur_pot1, **nfp2, **efp2, dtemp;

	// Node features
	set_node_potentials();

	// Node Extra features
	p_node = node_array + 1;
	nfp2 = node_extra_fea_para + 1;

	for (cur_item = 1; cur_item <= global_num_node; cur_item ++)
	{
		dtemp = *(*(nfp2 ++) + 1);
		
		cur_pot1 = p_node->node_pot + 2;
		*(cur_pot1 ++) += dtemp;
		*(cur_pot1 ++) += dtemp;

		dtemp *= 2;
		*(cur_pot1 ++) += dtemp;
		*(cur_pot1 ++) += dtemp;
		*cur_pot1 += dtemp;

		p_node ++;
	}

	//  Edge potentials
	p_edge = edge_array + 1;
	efp2 = edge_fea_para + 1;

	for (cur_item = 1; cur_item <= global_num_edge; cur_item ++)		// loop over edges
	{
		dtemp = edge_fea_intensity * (*efp2)[1];

		p_edge->edge_pot[1][3][5] += dtemp;
		p_edge->edge_pot[1][4][5] += dtemp;
		p_edge->edge_pot[1][6][2] += dtemp;
		p_edge->edge_pot[1][6][4] += dtemp;
		p_edge->edge_pot[1][6][5] += dtemp;

		p_edge->edge_pot[2][5][3] += dtemp;
		p_edge->edge_pot[2][5][4] += dtemp;
		p_edge->edge_pot[2][2][6] += dtemp;
		p_edge->edge_pot[2][4][6] += dtemp;
		p_edge->edge_pot[2][5][6] += dtemp;

		p_edge ++;	efp2 ++;
	}
}


char * CRFSensorChain::get_diagnosis(int argc, char ** argv)
{		
	return string_max_param;
}


}
