//#include "TreeBasics.hh"
#include <iostream>
#include <fstream>
#include <assert.h>
#include <vector>
#include <list>
#include <set>
#include <map>
#include <string>
#include <sys/timeb.h>
#include "Utility.hh"
using namespace std;

namespace tree_mcmc {


/**
 *	Extract an integer from arguments
 *	mode = 0 allows para not to exist in argv, and then return -1
 *	mode = 1 forces para to exist in argv, otherwise terminate
 *	mode = 2 when not exist, then add to argc, argv, and then return 0
 *	Return 1 if found the parameter, 0 if not found
 */
int Utility::get_arg(int *result, int *argc, char ** argv, char *para, int mode, int def_val)
{
	int i;

	for (i = 1; i < *argc; i++)
	{
		if (strcmp(argv[i], para) == 0)
		{
			*result = atoi(argv[i+1]);
			return 1;
		}
	}

	if (mode == 1)
	{
		cout<<"Missing parameter for "<<para<<endl;
		exit(1);
	}

	*result = def_val;

	if (mode == 2)
		add_arg(argc, argv, def_val, para);

	return 0;
}


/**
 *	Extract an unsigned integer from arguments
 *	mode = 0 allows para not to exist in argv, and then return -1
 *	mode = 1 forces para to exist in argv, otherwise terminate
 *	mode = 2 when not exist, then add to argc, argv, and then return 0
 *	Return 1 if found the parameter, 0 if not found
 */
int Utility::get_arg(unsigned int *result, int *argc, char ** argv, char *para, int mode, unsigned int def_val)
{
	int i;
	unsigned int temp;

	for (i = 1; i < *argc; i ++)
	{
		if (strcmp(argv[i], para) == 0)
		{
			temp = strtoul(argv[i+1], NULL, 10);
			*result = temp;
			return 1;
		}
	}

	if (mode == 1)
	{
		cout<<"Missing parameter for "<<para<<endl;
		exit(1);
	}

	*result = def_val;
	if (mode == 2)
		add_arg(argc, argv, def_val, para);

	return 0;
}


/**
 *	Extract a long integer from arguments
 *	mode = 0 allows para not to exist in argv, and then return -1
 *	mode = 1 forces para to exist in argv, otherwise terminate
 *	mode = 2 when not exist, then add to argc, argv, and then return 0
 *	Return 1 if found the parameter, 0 if not found
 */
int Utility::get_arg(long *result, int *argc, char ** argv, char *para, int mode, long def_val)
{
	int i;

	for(i = 1; i < *argc; i ++)
	{
		if (strcmp(argv[i], para) == 0)
		{
			*result = atol(argv[i+1]);
			return 1;
		}
	}

	if (mode == 1)
	{
		cout<<"Missing parameter for "<<para<<endl;
		exit(1);
	}

	*result = def_val;
	if (mode == 2)
		add_arg(argc, argv, def_val, para);

	return 0;
}


/**
 *	Extract a double precision floating point number from arguments
 *	mode = 0 allows para not to exist in argv, and then return -1
 *	mode = 1 forces para to exist in argv, otherwise terminate
 *	mode = 2 when not exist, then add to argc, argv, and then return 0
 *	Return 1 if found the parameter, 0 if not found
 */
int Utility::get_arg(double *result, int *argc, char ** argv, char *para, int mode, double def_val)
{
	int i;

	for(i = 1; i < *argc; i ++)
	{
		if (strcmp(argv[i], para) == 0)
		{
			*result = atof(argv[i+1]);
			return 1;
		}
	}
	if (mode == 1)
	{
		cout<<"Missing parameter for "<<para<<endl;
		exit(1);
	}

	*result = def_val;
	if (mode == 2)
		add_arg(argc, argv, def_val, para);
	return 0;
}


/**
 *	Extract an boolean from arguments
 *	mode = 0 allows para not to exist in argv, and then return -1
 *	mode = 1 forces para to exist in argv, otherwise terminate
 *	mode = 2 when not exist, then add to argc, argv, and then return 0
 *	Return 1 if found the parameter, 0 if not found
 */
int Utility::get_arg(bool *result, int *argc, char ** argv, char *para, int mode, bool def_val)
{
	int i;

	for(i = 1; i < *argc; i ++)
	{
		if (strcmp(argv[i], para) == 0)
		{
			*result = atoi(argv[i+1]);
			return 1;
		}
	}
	if (mode == 1)
	{
		cout<<"Missing parameter for "<<para<<endl;
		exit(1);
	}

	*result = def_val;
	if (mode == 2)
		add_arg(argc, argv, def_val, para);
	return 0;
}


/**
 *	Extract a string from arguments
 *	mode = 0 allows para not to exist in argv, and then return -1
 *	mode = 1 forces para to exist in argv, otherwise terminate
 *	mode = 2 when not exist, then add to argc, argv, and then return 0
 *	Return 1 if found the parameter, 0 if not found
 */
int Utility::get_arg(char **result, int *argc, char ** argv, char *para, int mode, char *def_val)
{
	int i;
	for (i = 1; i < *argc; i ++)
	{
		if(strcmp(argv[i], para) == 0)
		{
			if(*result == NULL)
				*result = new char[strlen(argv[i+1]) + 1];
			else 
			{
				delete [](*result);
				*result = new char[strlen(argv[i+1]) + 1];
			}

			strcpy(*result, argv[i+1]);
			return 1;
		}
	}

	if (mode == 1)
	{
		cout<<"Missing parameter for "<<para<<endl;
		exit(1);
	}

	if (*result != def_val)
	{
		if (def_val == NULL)
		{
			if (*result)
			{
				delete [] (*result);
				*result = NULL;
			}
		}
		else
		{
			delete [] (*result);
			*result = new char [strlen(def_val) + 1];
			strcpy(*result, def_val);
		}
	}
	if (def_val)
		add_arg(argc, argv, def_val, para);
	return 0;
}


void Utility::copy_arg(int * argc, char ** argv, int _argc, char ** _argv)
{
	int i;

	*argc = _argc;

	for (i = 0; i < _argc; i ++)
	{
		argv[i] = new char [strlen(_argv[i]) + 1];
		strcpy(argv[i], _argv[i]);
	}
}


/*
 *	Add an (flag, argument) pair for string to argc and argv
 */
void Utility::add_arg(int * argc, char ** argv, char * buffer, char *flag)
{
	int cur_arg;

	for (cur_arg = 1; cur_arg < *argc; cur_arg ++)
	{
		if (strcmp(argv[cur_arg], flag) == 0)		// if flag has been set in argv
		{
			cur_arg ++;
			delete []argv[cur_arg];
			argv[cur_arg] = new char [strlen(buffer) + 1];
			strcpy(argv[cur_arg], buffer);
			return;
		}
	}

	argv[*argc] = new char [strlen(flag) + 1];
	strcpy(argv[*argc], flag);
	(*argc) ++;	

	argv[*argc] = new char [strlen(buffer) + 1];
	strcpy(argv[*argc], buffer);
	(*argc) ++;
}


/*
 *	Add an (flag, argument) pair for integer to argc and argv
 */
void Utility::add_arg(int * argc, char ** argv, int value, char *flag)
{
	char buffer [32];
	sprintf(buffer, "%d", value);

	add_arg(argc, argv, buffer, flag);
}


/*
 *	Add an (flag, argument) pair for unsigned int to argc and argv
 */
void Utility::add_arg(int * argc, char ** argv, unsigned int value, char *flag)
{
	char buffer [32];
	sprintf(buffer, "%u", value);

	add_arg(argc, argv, buffer, flag);
}


/*
 *	Add an (flag, argument) pair for boolean to argc and argv
 */
void Utility::add_arg(int * argc, char ** argv, bool value, char *flag)
{
	char buffer [32];

	if (value)
		strcpy(buffer, "1");
	else
		strcpy(buffer, "0");

	add_arg(argc, argv, buffer, flag);
}


/*
 *	Add an (flag, argument) pair for long integer to argc and argv
 */
void Utility::add_arg(int * argc, char ** argv, long value, char *flag)
{
	char buffer [32];
	sprintf(buffer, "%ld", value);

	add_arg(argc, argv, buffer, flag);
}


/*
 *	Add an (flag, argument) pair for double precision floating point to argc and argv
 */
void Utility::add_arg(int * argc, char ** argv, double value, char *flag)
{
	char buffer [32];
	sprintf(buffer, "%lg", value);

	add_arg(argc, argv, buffer, flag);
}



void Utility::destroy_all_args(int argc, char **argv)
{
	for(int i = 0; i < argc; i ++)
		delete []argv[i];
	delete []argv;
}













}
