#ifndef Tree_mcmc_hh
#define Tree_mcmc_hh

//#include "Logsumexp.hh"
//#include "Utility.hh"

using namespace std;

namespace tree_mcmc {

// #define NINFINITY -1e307
// 
// typedef struct Node
// {
// 	int		sampled_value;	// The sampled value of this node in the last iteration
// 	int		num_class;		// How many classes/values can this node assume
// 	int		num_son;		// How many sons this node has
// 	int		max_num_son;	// Since Node space is shared by different trees, we need 
// 							//  the maximum possible number of sons in order to allocate space	
// 	double	*node_pot;		// Link to the node potential list
// 	double	*cond_pot;		// Link to the conditional potential list (has already multiplied by node potential)
// 	double	**msg_buffer;	// The buffer for all messages it has received [1][2]: [1] son_id, [2] label
// 	double	*sum_msg;		// sum of messages (for cache purposes)
// 	double	*rb_est;		// The RB estimation of the marginal probability
// } Node;
// 
// 
// typedef struct Edge
// {
// 	Node *p_node[3];			// Link to the two end nodes ptr1 < ptr2
// 	double **edge_pot[3];		// The edge potential tables (essentially the same). 
// 								//   [1] for i (row) < j (col), and [2] for i (row) > j (col)
// 	double **rb_est;			// The R-B estimation of edge potentials
// } Edge;
// 
// 
// typedef struct Tree 
// {
// 	/* 
// 	 *	Backward (sender, receiver) (forward just reverse order)
// 	 *	This forms a chain (sender, receiver) for forward, 
// 	 *   and a chain (sender, receiver) for backward (just reverse the order). 
// 	 */
// 	Node	**send, **rec;		
// 	int		*order_id;				// Backward the rank of son among sons (forward just reverse order)
// 	Edge	**tree_edge;			// Backward edge link (forward just reverse order)
// 	int		num_node;				// Number of nodes in this tree (forward just reverse order)
// 	int		*local_to_global;		// The id of the nodes in this tree
// 	int		*num_son;				// How many sons does each node has
// 	Edge	**cond_edge_array;		// for conditional use
// 	int		*cond_edge_index;		// [i] is the i^th node's starting conditional edge index
// 	bool	*cal_edge_rb;			// [i] denotes whether the i^th edge will contribute to the edge RB estimation.
// 	bool	cal_node_rb;			// whether to calculate the node RB estimates for this tree
// } Tree;
// 
// 
// typedef struct Partition
// {
// 	int num_trees;		// How many trees are there in the partition
// 	Tree *trees;	// The list of trees
// } Partition;


// extern int	global_num_node;		// Number of nodes
// 
// extern int	global_num_edge;		// Number of edges
// 
// extern int	global_num_partition;	// Number of partitions

//extern int	app_id;					// 1 for grid, 2 for sensor with supernodes, 3 for sensor with circle
extern int	topo_id;				// topology id
extern int	fea_id;					// feature type id
// extern int	**adj_matrix;			// Adjacency matrix 3
// 
// extern Node *node_array;			// array of the nodes
// extern Edge *edge_array;			// array of the edges
// 
// extern Partition	*partitions;	// all the partitions (all sets of trees)
// extern Tree			*init_tree;		// the tree for initialization
// extern double * sample_buf;			// sample buffer shared by all nodes
// extern double ** rb_buf;			// the buffer for edge RB.  Node RB uses sample_buf.

// extern double maxNodeRand;			// maximum random range of node feature parameters
// extern double maxEdgeRand;			// maximum random range of edge feature parameters
// 
// extern int	max_iter;				// maximum iteration (for sampling stop)
// extern int	max_time;				// maximum time (for sampling stop)

extern int	num_row;				// number of rows, only for grid
extern int	num_col;				// number of columns, only for grid
// extern int	num_class;				// number of classes (of each node)

extern int num_cross_upper;			// how many nodes are there above the horizontal chain?
extern int cross_hori_id;			// cross at which horizontal node?

extern int *current_phase;

extern int count_act1, count_act2;

// Input parameters
// extern int	num_obs_dim;			// number of node observation dimension
// extern int	num_edge_fea;			// number of edge features
// extern int	num_node_extra_fea;		// number of extra node feature

/*	Current parameters for the node features, [1][2][3]
 *		First index: node id;	second index: label;	third index: observation dimension
 */
// extern double	***node_fea_para;			
// 
// /*	Current parameters for the extra (non_NN) node features, [1][2]
//  *		First index: node id;	second index: extra feature index
//  */
// extern double	**node_extra_fea_para;
// 
// /*	Current parameters for the edge features, [1][2]
//  *		First index: edge id;	second index: feature index
//  */
// extern double	**edge_fea_para;
// 
// /*
//  *	The expectation of node features, [1][2][3]
//  *		First index: node id;	second index: label;	third index: observation/feature dimension
//  */
// extern double	***node_fea_exp;
// 
// /*
//  *	The expectation of node extra features, [1][2].  1st index: node id,  2nd index: feature index
//  */
// extern double	**node_extra_fea_exp;
// 
// /*
//  *	The expectation of edge features [1][2].  First index: edge id;	second index: feature index
//  */
// extern double	**edge_fea_exp;
// 	
// /*	The trace of the node features, [1][2][3]
//  *		First index: node id;	second index: label;	third index: observation/feature dimension
//  */
// extern double	***node_fea_trace;			
// 
// /*	The trace of the node extra features, [1][2]
//  *		First index: node id;	second index: feature index
//  */
// extern double	**node_extra_fea_trace;
// 
// /*	The trace of the edge features, [1][2]
//  *		First index: edge id;	second index: feature index
//  */
// extern double	**edge_fea_trace;			
// 
// /*	The current observation matrix, [1][2]
//  *		First index: node id;	second index: observation dimension index
//  */
// extern double	**cur_observation;


// The real value of edge features for CRF
extern double edge_fea_intensity;

// extern int cur_step;			// current step (how many steps have passed so far)
// extern int cycleTime;			// the cycle length
// extern int stepsPerControl;		// how many steps for each phase (used by bias)
// extern double bias;				// like CyclicBiasPolicy

extern bool use_local_reward;	// whether to use local reward or global reward
extern int max_delay;			// maximum delay to be considered in edge feature
extern int min_delay;			// minimum delay to be considered in edge feature
extern int **history_action;	// the history action of intersections (for edge feature use)

extern int edge_reward_def;		// -erd: edge reward definition
extern int edge_fea_sel;		// -efs: edge feature selection

// control of verbosity level (output of intermediate results)
// 1.  Display all the trees and init_tree
// 2.  Display the RB estimates of nodes after all iterations of tree MCMC sampling
// 4.  Display the RB estimates of edges after all iterations of tree MCMC sampling
// 8.  Display the node and edge potentials at the outset
// 16. Display number of iterations if max_time is used
// extern int verbosity;

extern char string_max_reward[1000];

extern double node_step_size;

extern bool exit_flag;

/************************************************************************/
/*     Functions begin                                                  */
/************************************************************************/

extern int get_edge_id(int node1, int node2);

/* All below are generic, not only for grid  */

// find a tree for initial sampling, result stored in init_tree
// Note: Construction of this tree requires discarding edges. 
// Only needs the adjacency matrix.
extern void build_init_tree(int root_id = -1);


// For 1D CRF (equivalent to define_partitions_grid + build_tree)
extern void build_ori_tree(int root_id = -1);


// Calculate the send, rec ... for the trees
// Similar to build_init_tree(), but different
// Note: Construction of this tree requires NO discarding of edges. 
extern void build_tree(Tree *cur_tree, int root_id = -1);


// Calculate the send, rec ... for all the trees in all the partitions
// This is generic, once the trees and partitions are defined.
// Let's skip the level of build_partition
extern void build_all_trees();


// clear the RB estimations
extern void clear_RB();


// Normalize the RB estimations
extern void normalize_RB();


// Sample from exp(log_prob[]), which has num_class elements
// Just sample, NO calculation of the RB estimation and store in sample_buf[]
// It Subtracts the maximum element.
extern int sample_distribution(double *log_prob, int num_class);


// Sample from exp(log_prob[]), which has num_class elements
// Also calculate the RB estimation and store in !!!sample_buf[]!!!
// It Subtracts the maximum element.
extern int infer_distribution(double *log_prob, int num_class);


// Calculate the conditional potentials of cur_tree
extern void cal_cond_pot(Tree *cur_tree);


// Perform forward message passing 
extern void forward_msg_pass(Tree *cur_tree);


// Get a sample of the root, run between forward msg prop and backward sampling
extern void sample_root(Tree *cur_tree);


// Get a RB estimation and sample of the root, run between forward and backward msg prop
extern void infer_root(Tree *cur_tree);


// Calculate the edge RB estimation
extern void fill_edge_rb(Edge *p_edge, Node *p_send, Node *p_rec, int order_id, int orient);

// Calculate the RB estimation of node features.
// Potentials are stored in p_node->sum_msg[]
extern void fill_node_rb(Node *p_node);


// Perform backward message passing (and sampling)
extern void backward_msg_pass(Tree *cur_tree);


// only backward sampling, no backward msg prop.
extern void backward_sample(Tree *cur_tree);


// Only draw a sample of the tree
extern void sample_tree(Tree *cur_tree);



// Perform inference on the whole grid using RB
extern void infer_general();

// Perform inference on the chain using RB on a CRF which is originally a tree
extern void infer_ori_tree();

// Will call infer_grid or infer_chain depending on graph_type
extern void infer_graph();



// putting everything together to sample a 2D CRF (only sample, no RB estimation)
extern void sample_general();

// putting everything together to sample CRF which is originally a tree (only sample, no RB estimation)
extern void sample_ori_tree();

// Will call sample_grid or sample_chain depending on graph_type
extern void sample_graph();



// After calculating the RB estimations on the graph,
// Calculate the feature expectations.
extern void cal_fea_expectation();


/************************************************************************/
/*     A lot of initializations (some are grid specific)                */
/************************************************************************/

extern void alloc_adj_matrix_grid();

extern void alloc_adj_matrix_chain();

extern void alloc_adj_matrix_circle();

extern void alloc_adj_matrix_cross();

extern void alloc_adj_matrix();

// allocate space for cur_observation, node_fea_para, edge_fea_para, node_fea_grad, edge_fea_grad
extern void alloc_fea();

extern void alloc_potentials();

extern void associate_edge_with_nodes_grid();

extern void associate_edge_with_nodes_chain();

extern void associate_edge_with_nodes_circle();

extern void associate_edge_with_nodes_cross();

extern void associate_edge_with_nodes();

// Fill up the node_fea_para[1][2][3]
//		First index: node id;	second index: label;	third index: observation 
// For debug only.
extern void set_node_fea_para_debug();


// Fill up the edge_fea_para[1][2]
//		First index: edge id;	second index: feature index
// For debug only.
extern void set_edge_fea_para_debug();


// Randomly initialize the node_fea_para[1][2][3]
//		First index: node id;	second index: label;	third index: observation 
extern void randomize_node_fea_para();


// Randomly initialize the edge_fea_para[1][2]
//		First index: edge id;	second index: feature index
extern void randomize_edge_fea_para();


// Set all trace cells to 0
extern void reset_trace();


// Discount the trace by factor discount
extern void discount_trace(double discount);


// Re-initialize randomly the feature parameters
extern void reset_params();


// Get the maximum feature parameter
extern double get_max_param();


// Write the graph parameter into a stream o
extern void graph_para_write(std::ostream& o);


// Read the graph parameter from a stream o
extern void graph_para_read(istream & o);


// For test.  Set the current observation matrix [1][2]
//		First index: node id;	second index: observation dimension index
// For debug only.
extern void set_cur_obs_debug();


// Add reward * trace to the feature parameters
// fea_para += reward * fea_trace
// Source: 1 add from trace, 2: add from grad (exp matrix)
extern void update_fea_para(double *stepSize, double *reward, int source);


// Add the gradient to the trace
// Implicitly fill the gradient into the exp[] space
extern void fill_trace();


// Put together the feature parameters and observations
extern void set_potentials();


extern void fill_edge_reward_grid(double *reward);


extern void fill_edge_reward_chain(double *reward);


extern void fill_edge_reward_generic(double *reward);


// construct trees in partitions
// determine the id of the nodes in the trees
extern void define_partitions_grid();

// construct trees in partitions
// determine the id of the nodes in the trees
extern void define_partitions_circle();


// set up the parameters, the adjacency matrix
extern void init_graph_chain(int *argc, char **argv);


// set up the parameters, the adjacency matrix
extern void init_graph_grid(int *argc, char **argv);


// set up the parameters, the adjacency matrix
extern void init_graph_circle(int *argc, char **argv);


// set up the parameters, the adjacency matrix
extern void init_graph(int *argc, char **argv);


// A generic sampler initializer
extern void init_tree_sampler(int *argc, char **argv);

// for delay based CRF (record the action taken in the previous step)
extern void record_previous_history();

// Dump the graph parameters
extern void dump_graph_specification(char * spec_file);


// Dump the inference results
extern void dump_infer_result(char * infer_file);


// Output the graph and inference results to file
extern void dump_graph(char * spec_file, char * infer_file);


// Bits in selecter: (from rightmost)
// 1.  Tree structure
// 2.  Node conditionals
// 4.  Node messages
// 8.  Node sum_msg
// 16. Samples drawn
extern void display_tree(Tree * cur_tree, unsigned selecter);


// Display the potentials of all nodes and edges
extern void display_pots();


// Display the RB estimates of the graph, including nodes and edges
extern void display_RB();

}

#endif
