#include <iostream>
#include <sstream>
#include <string>
#include <math.h>

#include "RLExtend.h"

#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/properties.hpp>
#include <boost/graph/subgraph.hpp>
#include <boost/random.hpp>
#include <boost/random/uniform_int.hpp>
#include <boost/random/uniform_real.hpp>
#include <boost/random/variate_generator.hpp>
#include <boost/timer.hpp>

#include <Generic.h>
#include <InputFunctions.h>
#include <Potential.h>
#include <Node.h>
#include <RandomVariable.h>
#include <OutputGraph.h>
#include <Experiment.h>

using namespace std;
using namespace boost;

double small;
double weight;

int		veryUgly;					// the last sampled value of the current node
int		verbose;					// verbosity level which controls console output verbosity

// Input parameters
double	***node_fea_para;			// current parameters for the node features, [1][2][3]
									//		first index: node id;	second index: action;	third index: observation dimension

double	**edge_fea_para;			// current parameters for the edge features, [1][2]
									//		first index: edge id;	second index: feature index

extern double	**edge_fea_exp;		// The expectation of edge features [1][2].  
									//		First index: edge id;	second index: feature index

extern double	***node_fea_exp;	// The expectation of node features, [1][2].  
									//		1st index: node id,  2nd index: input dim id

double	***node_fea_trace;			// the trace of the node features, [1][2][3]
									//		first index: node id;	second index: action;	third index: observation dimension

double	**edge_fea_trace;			// the trace of the edge features, [1][2]
									//		first index: edge id;	second index: feature index


double	**cur_observation;			// the current observation matrix, [1][2]
									//		first index: node id;	second index: observation dimension index

//	Sampling and expectation evaluator work variables

//int		*time_edge_visit;		// how many times an edge has been covered, shared by all edge features, [1]
									//		index: edge id
									// shared by all the edge features of a specific edge

//int		*time_node_visit;		// how many times this node feature estimator has been evaluated, [1]
									//		index: node id
									// shared by all the node features of the specific node

double	***edge_margin_record;		// the accumulated edge RB estimation, [1][2][3]
									//		first index: edge id;	second index: action id of source; 
									//		third index: action id of target
									// only record the #action * #action matrix, and all features can be calculated based on it

double	**node_margin_record;		// the accumulated node RB estimation, [1][2]
									//		first index: node id;	second index: action id;  
 									// only record the #action vector, and all features can be calculated based on it

unsigned int		*sample_record;				// The sample result of the last run of sampler.

double	**edge_fea_exp;					// The expectation of edge features

double	***node_fea_exp;					// The expectation of node features, [1][2]
									//		1st index: node id,  2nd index: input dim id

//	Auxiliary data structure, all such variables end with _aux
double	**node_potential_aux;		// node potentials to calculate the expectations of node features, 
									//		first index: node id;	second index: action

double	**edge_potential_aux;		//		For each edge, store the potential of each pair of configuration
									//		Used for normalization, and purged when changing to a new edge.
									//		To avoid repeated allocation of memory for computing edge potential ().
									//		first index:	action of the source node of the edge;	
									//		second index:	action of the target node of the edge;	

int * node_assign;					// Record for each edge, what nodes' potential will be 
									//		incorporated into this edges' potential
									// node_assign[i] contains the node id whose potential is incorporated into edge i.
									// node_assign[i] = -1 iff the there is no node whose potential is incorporated into node i.

vector < double> ** edge_potential_table_crf;	// The edge potential table
											// [1]->: first index is the edge id, then get the pointer to the vector <double>

vector < double> ** node_potential_table_crf;	// The node potential table
											// [1]->: first index is the node id, then get the pointer to the vector <double>


//	Scalar parameters
unsigned int	dim_horizontal;		// grid horizontal size (number of nodes, not number of horizontal edges!!!)
unsigned int	dim_vertical;		// grid vertical size (number of nodes, not number of horizontal edges!!!)
unsigned int	dim_observation;	// how many sources of observation for each node
unsigned int	dim_action;			// how many actions each agent can take
unsigned int	num_nodes;			// number of nodes = dim_horizontal * dim_vertical
unsigned int	num_edge_max;		// number of edges + dim_vertical = 2 * num_nodes - dim_dim_horizontal
unsigned int	num_edge_fea;		// number of edge features (preliminarily set to 1, indicator function)


// Value in the constructor is the seed...

lagged_fibonacci607 fibonnacci_number_generator (20);
timer global_timer;

typedef property< edge_index_t, unsigned int, ContinuousPotential *> NPEdgeProperty;
typedef adjacency_list<vecS, vecS, undirectedS, GraphicalModelNode *, 	NPEdgeProperty > NPGraph;

typedef boost::property< boost::edge_index_t, unsigned int, DiscretePotential *> DiscretePairwiseEdgeProperty;
typedef boost::adjacency_list<boost::vecS, boost::vecS, boost::undirectedS, MessageNode<DiscreteRandomVariable, boost::adjacency_list_traits<boost::vecS, boost::vecS, boost::undirectedS>::vertex_descriptor> *, 	DiscretePairwiseEdgeProperty > DiscretePairwiseGraph;

Experiment * current_experiment;

bool monitoring = false;

//fibonnacci_number_generator.seed(time(NULL));

variate_generator <lagged_fibonacci607 &, uniform_real <  > > uniform_distribution_0_1 (fibonnacci_number_generator, uniform_real < > (0.0, 1.0) );

unsigned int mrf_rows = 10;
unsigned int mrf_columns = 10;


int main_PGM_from_file(std::string & input)
{
	Experiment *main_experiment;
	
	main_experiment = new Experiment;

	if (!main_experiment->set_up_model_from_file(input))
	{
		cout << "A fatal error occurred while parsing file. Aborting..." << endl;
		return 1;
	}
	
	main_experiment->get_inference();
	
//	main_experiment.export_results_to_file( argv[2]);
	
    return 0;
}


int main_PGM_string(stringstream & input)
{
	Experiment *main_experiment;
	
	main_experiment = new Experiment;

	if (!main_experiment->set_up_model_from_string(input))
	{
		cout << "A fatal error occurred while parsing file. Aborting..." << endl;
		return 1;
	}
	
	main_experiment->get_inference();
	
//	main_experiment.export_results_to_file( argv[2]);
	
	delete main_experiment;
	
	prepare_string_input(input, 10, NULL);
    return 0;
}

/******************************************************************************/
/*		Xinhua's utilities that parses the input switches and arguments		  */
/******************************************************************************/


// extract an integer for para (eg, -a, -t)
// mode = 1 forces para to exist in argv, otherwise terminate
// mode = 0 allows para not to exist in argv, and then return -1
// return 1 if found the parameter, 0 if not found
int getParaInt(int *result, int argc, char * const argv[], char *para, int mode)
{
	int i;
	for(i=1; i<argc; i++)
	{
		if(strcmp(argv[i], para) == 0)
		{
			*result = atoi(argv[i+1]);
			return 1;
		}
	}

	if(mode == 1)
	{
		cout<<"Missing parameter for "<<para<<endl;
		exit(1);
	}

	return 0;
}


// extract an integer for para (eg, -a, -t)
// mode = 1 forces para to exist in argv, otherwise terminate
// mode = 0 allows para not to exist in argv, and then return -1
// return 1 if found the parameter, 0 if not found
int getParaUnInt(unsigned int *result, int argc, char * const argv[], char *para, int mode)
{
	int i;
	unsigned int temp;
	for(i=1; i<argc; i++)
	{
		if(strcmp(argv[i], para) == 0)
		{
			temp = strtoul(argv[i+1], NULL, 10);
			*result = temp;
			return 1;
		}
	}

	if(mode == 1)
	{
		cout<<"Missing parameter for "<<para<<endl;
		exit(1);
	}

	return 0;
}


// extract an integer for para (eg, -a, -t)
// mode = 1 forces para to exist in argv, otherwise terminate
// mode = 0 allows para not to exist in argv, and then return -1
// return 1 if found the parameter, 0 if not found
int getParaLong(long *result, int argc, char * const argv[], char *para, int mode)
{
	int i;
	for(i=1; i<argc; i++)
	{
		if(strcmp(argv[i], para) == 0)
		{
			*result = atol(argv[i+1]);
			return 1;
		}
	}

	if(mode == 1)
	{
		cout<<"Missing parameter for "<<para<<endl;
		exit(1);
	}

	return 0;
}


// extract an floating point number for para (eg, -r, -c)
// mode = 1 forces para to exist in argv, otherwise terminate
// mode = 0 allows para not to exist in argv, and then return -1
// return 1 if found the parameter, 0 if not found
int getParaDouble(double *result, int argc, char * const argv[], char *para, int mode)
{
	int i;
	for(i=1; i<argc; i++)
	{
		if(strcmp(argv[i], para) == 0)
		{
			*result = atof(argv[i+1]);
			return 1;
		}
	}
	if(mode == 1)
	{
		cout<<"Missing parameter for "<<para<<endl;
		exit(1);
	}

	return 0;
}


// extract an floating point number for para (NOT used in this project)
// mode = 1 forces para to exist in argv, otherwise terminate
// mode = 0 allows para not to exist in argv, and then return -1
int getParaString(char **result, int argc, char * const argv[], char *para, int mode)
{
	int i;
	for(i=1; i<argc; i++)
	{
		if(strcmp(argv[i], para) == 0)
		{
			if(*result == NULL)
				*result = new char[strlen(argv[i+1]) + 1];
			else 
			{
				delete [](*result);
				*result = new char[strlen(argv[i+1]) + 1];
			}

			strcpy(*result, argv[i+1]);
			return 1;
		}
	}

	if(mode == 1)
	{
		cout<<"Missing parameter for "<<para<<endl;
		exit(1);
	}

	return -1;
}

/*************************************************************************************/
/*		End of Xinhua's utilities that parses the input switches and arguments		 */
/*************************************************************************************/



/*************************************************************************************/
/*		Common utilities for preparing the sampling									 */
/*************************************************************************************/


/**
 *	Prepare the input of the network, with special treatment for CRF, 
 *	i.e., we do not provide the network parameter in the normal way
 *	We will hard code the configurations by some special functions
 */
void prepare_string_input(std::stringstream & input, int argc, char** argv)
{
	unsigned i;	
	int max_time, max_loop, limit_mode;

	// first check whether timer or max_loop is selected
	max_time = -1;
	max_loop = -1;
	limit_mode = 1;		// 1 for timer, 2 for max_loop

	getParaInt(&max_time, argc, argv, "-tm", 0);
	getParaInt(&max_loop, argc, argv, "-lp", 0);

	if(max_time < 0 && max_loop < 0)
	{
		cout<<"Neither timer nor loop limit is set.  Choosing to use timer for 100 milliseconds"<<endl;
		max_time = 100;
	}
	if(max_time > 0 && max_loop > 0)
	{
		cout<<"Both timer and loop limit are set.  Will use timer."<<endl;
	}
	if(max_time < 0 && max_loop > 0)
		limit_mode = 2;

	input << "<method> (tree_mcmc)( ";
	if(limit_mode == 1)
		input << max_time <<" ) \n";
	else if(limit_mode == 2)
		input << max_loop <<" ) \n";

	input << "<number of vertices> ( " << num_nodes<<" )\n";
	input << "<model type> (pairwise-discrete) \n";
	
	input << "<random variable sizes> ";
	for(i = 1; i <= num_nodes; i++)
		input << "( "<<i<<" "<<dim_action<<" ) ";
	input << "\n";

//	input << "<all random variable sizes> ( "<<dim_action<<" )\n";
	
	input << "<link crf> ("<<dim_horizontal<<" "<<dim_vertical<<") \n";
	input << "<potential tables crf>\n";
	
//	input << "<observations> ";
//	for(i = num_nodes + 1; i <= total; i ++)
//		input << "(" << i << " 0) ";
//	input << "("<<num_nodes + 1<<" 0) ";
//	input << "\n";

	input << "<internal potentials crf>\n";

	if(limit_mode == 1)	
		input << "<timer>\n";
}

/**
 *	Print the marginals for all edges (mainly for debug purposes).
 */
void print_edge_marginals()
{
	unsigned int i, j, k, loop_flag, cur_edge;
	double sum;

	loop_flag = 0;
	for (cur_edge = 0; cur_edge < num_edge_max; cur_edge++)
	{		
		if(cur_edge < num_nodes)
		{
			loop_flag++;
			if(loop_flag == dim_horizontal)		// skip non-existent features
			{
				loop_flag = 0;
				continue;
			}
		}

		if(cur_edge == num_nodes)
			cout<<"||||||||||||||||||||||||||||||||||||||||||||||||||||||"<<endl;

		sum = 0;
		for(i = 0; i < dim_action; i ++)
		{
			for (j = 0; j < dim_action; j++)
			{
				sum += edge_margin_record[cur_edge][i][j];
			}
		}

		cout<<"== Edge: "<<cur_edge<<" =="<<endl;

		for(i = 0; i < dim_action; i ++)
		{
			for (j = 0; j < dim_action; j++)
			{
				cout<<edge_margin_record[cur_edge][i][j] / sum<<"\t";				
			}
			cout<<endl;
		}
	}
}


/**
 *	This function is already useless.  But wasted me a lot of time so I still leave it here.
 *	Initialize node_assign[].  To be called only once in the RL iteration.
 */
void initialize_assign_node_to_edges()
{
	unsigned int cur_node, i, top, bottom, left, right, top_left, cur_edge;

	node_assign = new int [num_edge_max];
	for(i = 0; i < num_edge_max; i++)
		node_assign[i] = -1;
	
	top = left = 0;
	right = dim_horizontal - 1;
	bottom = dim_vertical - 1;

	while(true)
	{
		if(top >= bottom || left >= right)
			break;

		top_left = top * dim_horizontal + left;

		// first, the top row
		cur_node = top_left;						// initialize to the top-left corner
		for(i = left; i < right; i++, cur_node++)
		{
			node_assign[cur_node] = cur_node + 1; 
		}

		// second, the bottom row
		cur_node = bottom * dim_horizontal + left;	// initialize to the bottom-left corner
		for(i = left; i < right; i++, cur_node++)
		{
			node_assign[cur_node] = cur_node;
		}

		// third, the left column
		cur_node = top_left;						// initialize to the top-left corner
		cur_edge = cur_node + num_nodes;
		for(i = top; i < bottom; i++)
		{
			node_assign[cur_edge] = cur_node; 
			cur_node += dim_horizontal;
			cur_edge += dim_horizontal;
		}

		// fourth, the right column
		cur_node = top_left + right - left;			// initialize to the top-right corner
		cur_edge = cur_node + num_nodes;
		for(i = top; i < bottom; i++)
		{
			node_assign[cur_edge] = cur_node + dim_horizontal;
			cur_node += dim_horizontal;
			cur_edge += dim_horizontal;
		}

		top ++; 		left ++;
		right --;		bottom --;
	}

	if(top > bottom || left > right)
	{
//		return;
	}
	else if(top == bottom)		// there leaves a row
	{
		cur_node = top * dim_horizontal + left - 1;			// initialize to the left of the leftmost node 
		assert(dim_vertical > 1);

		for(i = left - 1; i < right; i++, cur_node++)
		{
			node_assign[cur_node] = cur_node + 1;
		}			
	}
	else		//	if(left == right)	there leaves a column
	{
		assert(dim_horizontal > 1);

		cur_node = (top - 1) * dim_horizontal + left;			// initialize to the above of the top node
		cur_edge = cur_node + num_nodes;
		for(i = top - 1; i < bottom; i++)
		{
			node_assign[cur_edge] = cur_node + dim_horizontal;
			cur_node += dim_horizontal;
			cur_edge += dim_horizontal;
		}
	}

// 	for(i = 0; i < num_edge_max; i++)
// 		cout<<i<<": "<<node_assign[i]<<endl;
}


/*
 *	double	***edge_margin_record;	// the accumulated edge RB estimation, [1][2][3]
 *									//		first index: edge id;	second index: action id of source; 
 *									//		third index: action id of target
 *									// only record the #action * #action matrix, and all features can be calculated based on it
 *	Clear all items into 0
 */
void clear_edge_margin_record()
{
	unsigned i, j, k, l;
	double ***cur_edge, *cur_1, **cur_2;

	cur_edge = edge_margin_record;

	// horizontal edges
	for(i = 0; i < dim_vertical; i++)
	{
		for(j = 0; j < dim_horizontal - 1; j++)
		{
			cur_2 = *cur_edge;
			for(k = 0; k < dim_action; k++)
			{				
				cur_1 = *(cur_2++);
				for (l = 0; l < dim_action; l ++)
				{
					*(cur_1 ++) = 0;
				}
			}

			cur_edge++;
		}
		cur_edge++;
	}

	// vertical edges
	for(i = dim_vertical * dim_horizontal; i < num_edge_max; i++, cur_edge++)
	{
		cur_2 = *cur_edge;

		for(k = 0; k < dim_action; k++)
		{
			cur_1 = *(cur_2++);
			for (l = 0; l < dim_action; l ++)
			{
				*(cur_1 ++) = 0;
			}
		}
	}
}


/*
 *	double	**node_margin_record;		// the accumulated node RB estimation, [1][2]
 *									//		first index: node id;	second index: action id;  
 *									// only record the #action vector, and all features can be calculated based on it
 *	Clear all elements into 0
 */
void clear_node_margin_record()
{
	unsigned i, j;
	double **cur_node2, *cur_node1;

	cur_node2 = node_margin_record;
	for(i = 0; i < num_nodes; i++)
	{
		cur_node1 = *(cur_node2++);
		for(j = 0; j < dim_action; j++)
			*(cur_node1++) = 0;
	}
}


/*
 *	double	***node_fea_grad;		// the grad/trace of node features, [1][2][3]
 *								1st index: node id;	2nd index: action;	3rd index: observation dimension
 *	Clear all elements into 0
 */
void clear_node_grad()
{

	unsigned int i, j, k;
	double *** cur_3, ** cur_2, * cur_1;

	cur_3 = node_fea_grad;

	for(i = 0; i < num_nodes; i++)
	{
		cur_2 = *(cur_3++);

		for(j = 0; j < dim_action; j++)
		{			
			cur_1 = *(cur_2++);

			for(k = 0; k < dim_observation; k++)
			{
				*(cur_1++) = 0;
			}
		}
	}

}


/*
 *	double	**edge_fea_grad;		// the grad/trace of edge features, [1][2]
 *											1st index: edge id;	2nd index: feature index
 *	Clear all elements into 0
 */
void clear_edge_grad()
{
	
	unsigned int i, j, k;
	double *** cur_3, ** cur_2, * cur_1;
	
	cur_2 = edge_fea_grad;

	// First, horizontal edges
	for(i = 0; i < dim_vertical; i++)
	{
		for(j = 0; j < dim_horizontal - 1; j++)
		{
			cur_1 = *(cur_2++);

			for(k = 0; k < num_edge_fea; k++)
			{
				*(cur_1++) = 0;
			}
		}
		cur_2++;
	}

	// Second, vertical edges
	for(i = dim_vertical * dim_horizontal; i < num_edge_max; i++)
	{
		cur_1 = *(cur_2++);

		for(k = 0; k < num_edge_fea; k++)
		{
			*(cur_1++) = 0;
		}
	}	
}


/*************************************************************************************/
/*		End of Common utilities for preparing the sampling							 */
/*************************************************************************************/



/*************************************************************************************/
/*		Initializations (allocate spaces), common to all							 */
/*************************************************************************************/


/** 
 *	It allocates the VECTOR space for edge_potential_table_crf.  
 *	To be called only once in the RL iteration.
 *	Later, just change the value in the vector, no re-allocate.
 *	Also calculates the node_assign[] (to which edge each node potential is assigned to).
 *	This function does not calculate any edge potential value.  It just allocates the space.
 */
void initialize_edge_potential()
{
	unsigned int cur_edge, total, loop_flag;

//	initialize_assign_node_to_edges();
	edge_potential_table_crf = new vector <double> * [num_edge_max];

	loop_flag = 0;
	total = dim_action * dim_action;
	for (cur_edge = 0; cur_edge < num_edge_max; cur_edge++)
	{		
		if(cur_edge < num_nodes)
		{
			loop_flag++;
			if(loop_flag == dim_horizontal)		// skip non-existent features
			{
				loop_flag = 0;
				edge_potential_table_crf[cur_edge] = NULL;
				continue;
			}
		}

		edge_potential_table_crf[cur_edge] = new vector <double> (total);
	}

// 	for (; cur_edge < num_edge_max + num_nodes; cur_edge++)	 // for edges connected to observation nodes
// 	{
// 		edge_potential_table_crf[cur_edge] = new vector <double> (dim_action);
//	}
}


/* 
 *	It allocates the VECTOR space for node_potential_table_crf.
 *	To be called only once in the RL iteration.
 *	Later, just change the value in the vector, no re-allocate.
 *	This function does not calculate any node potential value.  It just allocates the space.
 */
void initialize_node_potential()
{
	unsigned int cur_node, total, loop_flag;

	node_potential_table_crf = new vector <double> * [num_nodes];

	for (cur_node = 0; cur_node < num_nodes; cur_node++)
	{		
		node_potential_table_crf[cur_node] = new vector <double> (dim_action);
	}
}


/**
 *	run only once in the very beginning
 *	double	**edge_fea_para;			// current parameters for the edge features, [1][2]
 *										//		first index: edge id;	second index: feature index
 *	We do not initialize the terms into any value.  Initialization will be handled by update_edge_fea_para or randomize_edge_fea_para
 */
void initialize_edge_fea_para()
{
	unsigned i, j;
	double **cur_edge, *cur_1;
	double **cur_edge_grad, *cur_1_grad;
	double **cur_edge_exp, *cur_1_exp;

	edge_fea_para = new double * [num_edge_max];
	edge_fea_grad = new double * [num_edge_max];
	edge_fea_exp = new double * [num_edge_max];
	
	cur_edge = edge_fea_para;
	cur_edge_grad = edge_fea_grad;
	cur_edge_exp = edge_fea_exp;

	// horizontal edges
	for(i = 0; i < dim_vertical; i++)
	{
		for(j = 0; j < dim_horizontal - 1; j++)
		{
			*(cur_edge++) = new double [num_edge_fea];			// edge_fea_para[i*dim_horizontal+j] = new []
			*(cur_edge_grad++) = new double [num_edge_fea];		// edge_fea_grad[i*dim_horizontal+j] = new []
			*(cur_edge_exp++) = new double [num_edge_fea];		// edge_fea_exp[i*dim_horizontal+j] = new []
		}	
		cur_edge++; cur_edge_grad++;  cur_edge_exp++;
	}

	// vertical edges
	for(i = dim_vertical * dim_horizontal; i < num_edge_max; i++)
	{
		*(cur_edge++) = new double [num_edge_fea];				// edge_fea_para[i] = new []
		*(cur_edge_grad++) = new double [num_edge_fea];			// edge_fea_grad[i] = new []
		*(cur_edge_exp++) = new double [num_edge_fea];			// edge_fea_exp[i] = new []
	}
}


/**
 *	run only once in the very beginning
 *	***node_fea_para:  current parameters for the node features, [1][2][3]
 *						first index: node id;	second index: action;	third index: observation dimension
 *	We do not initialize the terms into any value.  Initialization will be handled by update_node_fea_para or randomize_node_fea_para
 */
void initialize_node_fea_para()
{
	unsigned int i, j;
	double *** cur_3, ** cur_2;
	double *** cur_3_grad, ** cur_2_grad;
	double *** cur_3_exp, ** cur_2_exp;

	node_fea_para = new double **[num_nodes];
	node_fea_grad = new double **[num_nodes];
	node_fea_exp = new double **[num_nodes];

	cur_3 = node_fea_para;
	cur_3_grad = node_fea_grad;
	cur_3_exp = node_fea_exp;

	for(i = 0; i < num_nodes; i++)
	{
		cur_2 = new double *[dim_action];
		cur_2_grad = new double *[dim_action];
		cur_2_exp = new double *[dim_action];

		*(cur_3++) = cur_2; 
		*(cur_3_grad++) = cur_2_grad; 
		*(cur_3_exp++) = cur_2_exp; 
		
		for(j = 0; j < dim_action; j++)
		{
			*(cur_2++) = new double [dim_observation];			// node_fea_para[i][j] = new double []
			*(cur_2_grad++) = new double [dim_observation];		// node_fea_grad[i][j] = new double []
			*(cur_2_exp++) = new double [dim_observation];		// node_fea_exp[i][j] = new double []
		}
	}
}


/**
 *	double	***edge_margin_record;		// the accumulated edge RB estimation, [1][2][3]
 *									//		first index: edge id;	second index: action id of source; 
 *									//		third index: action id of target
 *									// only record the #action * #action matrix, and all features can be calculated based on it
 */
void initialize_edge_margin_record()
{
	unsigned i, j, k;
	double ***cur_edge, **cur_2;

	edge_margin_record = new double ** [num_edge_max];
	cur_edge = edge_margin_record;

	// horizontal edges
	for(i = 0; i < dim_vertical; i++)
	{
		for(j = 0; j < dim_horizontal - 1; j++)
		{
			cur_2 = new double * [dim_action];
			*cur_edge = cur_2;
			for(k = 0; k < dim_action; k++)
			{
				*(cur_2++) = new double [dim_action];
			}

			cur_edge++;
		}
		cur_edge++;
	}

	// vertical edges
	for(i = num_nodes; i < num_edge_max; i++, cur_edge++)
	{
		cur_2 = new double * [dim_action];
		*cur_edge = cur_2;

		for(k = 0; k < dim_action; k++)
		{
			*(cur_2++) = new double [dim_action];
		}
	}

	// observation edges
// 	for(i = 0; i < num_nodes; i++, cur_edge++)
// 	{
// 		cur_2 = new double * [dim_action];
// 		*cur_edge = cur_2;
// 
// 		for(k = 0; k < dim_action; k++)
// 		{
// 			*(cur_2++) = new double [dim_action];
// 		}
//	}	
}


/**
 *	double	**node_margin_record;	the accumulated node RB estimation, [1][2]
 *										first index: node id;	second index: action id;  
 *										only record the #action vector, and all features can be calculated based on it
 */
void initialize_node_margin_record()
{
	unsigned i;
	double **cur_node;

	node_margin_record = new double * [num_nodes];
	cur_node = node_margin_record;

	for(i = 0; i < num_nodes; i++)
	{
		*(cur_node++) = new double[dim_action];
	}
}


/**
 *	double	*edge_potential_aux;	// just to avoid repeated allocation of memory for computing edge potential (normalization)
 *									//		first index:	action of the source node of the edge;	
 *									//		second index:	action of the target node of the edge;	
 */
void initialize_edge_potential_aux()
{
	unsigned int i;
	double **cur_2;

	edge_potential_aux = new double * [dim_action];
	cur_2 = edge_potential_aux;

	for(i = 0; i < num_nodes; i++)
		*(cur_2 ++) = new double [dim_action];
}


/**
 *	double	**node_potential_aux;	// node potentials to calculate the expectations of node features, 
 *									//		first index: node id;	second index: action
 */
void initialize_node_potential_aux()
{
	unsigned int i;
	double **cur_2;

	node_potential_aux = new double * [num_nodes];
	cur_2 = node_potential_aux;

	for(i = 0; i < num_nodes; i++)
		*(cur_2 ++) = new double [dim_action];
}


/* 
 *	Initialize the current observation matrix,   [1][2]
 *		1st index: node id;		2nd index: observation dimension index
 *	Allocates the memory.  Will be run only once.  Not assign any value.
 */
void initialize_observation()
{
	unsigned int i;
	double **cur_2;

	cur_observation = new double *[num_nodes];
	cur_2 = cur_observation;

	for (i = 0; i < num_nodes; i++)
	{
		*(cur_2++) = new double [dim_observation];
	}	
}


/*************************************************************************************/
/*		End of initializations (allocate spaces), common to all						 */
/*************************************************************************************/



/*************************************************************************************/
/*		Updates feature parameters, common to all									 */
/*************************************************************************************/


/**
 *	run at each iteration/update of RL
 *	double	**edge_fea_para;			// current parameters for the edge features, [1][2]
 *										//		first index: edge id;	second index: feature index
 */
void update_edge_fea_para(double step_size)
{
	unsigned i, j, k;
	double **cur_edge, *cur_1;
	double **cur_grad, *cur_grad_1;

	cur_edge = edge_fea_para;
	cur_grad = edge_fea_grad;

	// horizontal edges
	for(i = 0; i < dim_vertical; i++)
	{
		for(j = 0; j < dim_horizontal - 1; j++)
		{
			cur_1 = *cur_edge;
			cur_grad_1 = *cur_grad;

			for(k = 0; k < num_edge_fea; k++)
			{
				*(cur_1++) += *(cur_grad_1++) * step_size;			
				// update the horizontal edge parameters, edge_fea_para[i*dim_horizontal+j][k]
			}
			cur_edge++;
			cur_grad++;
		}
		cur_edge++;
		cur_grad++;
	}

	// vertical edges
	for(i = dim_vertical * dim_horizontal; i < num_edge_max; i++)
	{
		cur_1 = *(cur_edge++);
		cur_grad_1 = *(cur_grad++);

		for(k = 0; k < num_edge_fea; k++)
		{
			*(cur_1++) += *(cur_grad_1++) * step_size;				
			// change here to update the vertical edge parameters, edge_fea_para[i][k]
		}
	}
}


/**
 *	run at each iteration/update of RL
 *	***node_fea_para:  current parameters for the node features, [1][2][3]
 *						first index: node id;	second index: action;	third index: input dimension
 */
void update_node_fea_para(double step_size)
{
	unsigned int i, j, k;
	double *** cur_3, ** cur_2, * cur_1;
	double *** grad_cur_3, ** grad_cur_2, * grad_cur_1;

	cur_3 = node_fea_para;
	grad_cur_3 = node_fea_grad;

	for(i = 0; i < num_nodes; i++)
	{
		cur_2 = *(cur_3++);
		grad_cur_2 = *(grad_cur_3++);


		for(j = 0; j < dim_action; j++)
		{			
			cur_1 = *(cur_2++);
			grad_cur_1 = *(grad_cur_2++);

			for(k = 0; k < dim_observation; k++)
			{
				*(cur_1++) += *(grad_cur_1++) * step_size;	
				// update node feature parameter here, this is actually node_fea_para[i][j][k]
			}
		}
	}
}


/*************************************************************************************/
/*		End of updates feature parameters, common to all							 */
/*************************************************************************************/




/*************************************************************************************/
/*		Randomize feature parameters, common to all									 */
/*************************************************************************************/


/*
 *	run at the beginning of RL, to randomly initialize the edge feature parameters
 *	double	**edge_fea_para;			// current parameters for the edge features, [1][2]
 *										//		first index: edge id;	second index: feature index
 */
void randomize_edge_fea_para(double maxRand)
{
	unsigned i, j, k;
	double **cur_edge, *cur_1;

	typedef boost::minstd_rand base_generator_type;
	base_generator_type generator(42u);
	boost::uniform_real<> uni_dist(- maxRand, maxRand);
	boost::variate_generator<base_generator_type&, boost::uniform_real<> > rand_gen(generator, uni_dist);

	cur_edge = edge_fea_para;

	// First, horizontal edge parameters
	for(i = 0; i < dim_vertical; i++)
	{
		for(j = 0; j < dim_horizontal - 1; j++)
		{
			cur_1 = *cur_edge;
			for(k = 0; k < num_edge_fea; k++)
			{
				*(cur_1++) = rand_gen();		// edge_fea_para[i*dim_horizontal+j][k]
			}
			cur_edge++;
		}
		cur_edge++;
	}

	// Second, vertical edge parameters
	for(i = dim_vertical * dim_horizontal; i < num_edge_max; i++)
	{
		cur_1 = *(cur_edge++);
		for(k = 0; k < num_edge_fea; k++)
		{
			*(cur_1++) = rand_gen();		// edge_fea_para[i][k]
		}
	}
}


/*
 *	run at the beginning of RL, to randomly initialize the edge feature parameters
 *	***node_fea_para:  current parameters for the node features, [1][2][3]
 *						first index: node id;	second index: action;	third index: input dimension
 */
void randomize_node_fea_para(double maxRand)
{
	unsigned int i, j, k;
	double *** cur_3, ** cur_2, * cur_1;

	typedef boost::minstd_rand base_generator_type;
	base_generator_type generator(42u);
	boost::uniform_real<> uni_dist(- maxRand, maxRand);
	boost::variate_generator<base_generator_type&, boost::uniform_real<> > rand_gen(generator, uni_dist);

	cur_3 = node_fea_para;

	for(i = 0; i < num_nodes; i++, cur_3++)
	{
		cur_2 = *cur_3;

		for(j = 0; j < dim_action; j++, cur_2 ++)
		{			
			cur_1 = *cur_2;
			for(k = 0; k < dim_observation; k++, cur_1++)
			{				
				*cur_1 = rand_gen();				// this is actually node_fea_para[i][j][k]
			}
		}
	}
}


/* 
 *	Randomize the current observation matrix,   [1][2]
 *		1st index: node id;		2nd index: observation dimension index
 *	Will be run only once.
 */
void randomize_observation(double maxRand)
{
	unsigned int i, j;
	double **cur_2, *cur_1;

	typedef boost::minstd_rand base_generator_type;
	base_generator_type generator(42u);
	boost::uniform_real<> uni_dist(- maxRand, maxRand);
	boost::variate_generator<base_generator_type&, boost::uniform_real<> > rand_gen(generator, uni_dist);

	cur_2 = cur_observation;

	for (i = 0; i < num_nodes; i++)
	{		
		cur_1 = * (cur_2++);
		
		for (j = 0; j < dim_observation; j++)
		{
			*(cur_1 ++) = rand_gen();			// actually cur_observation[i][j]
		}
	}	
}


/*************************************************************************************/
/*		End of randomize feature parameters, common to all							 */
/*************************************************************************************/




/****************************************************************************/
/*	NOTE:																	*/
/*  The initialization/update/randomize of feature PARAMETERS are general	*/
/*  The update of potentials is for grid specifically.						*/
/****************************************************************************/




/*************************************************************************************/
/*		Update potentials for Grid													 */
/*************************************************************************************/



/*
 *	Update the potential table of the edges, to be called in each iteration of RL.
 *	It does NOT reallocate the VECTOR space for edge_potential_table_crf.  Just change the value in the vector.
 *	Directly use the node_assign array of the node potentials (which does not change in RL iteration)
 *	This function is very application specific.  For different tasks, one may want to re-write this function.
 *	Now it is written for 2D CRF with NN node features (linear)
 */
void update_edge_potential_grid()
{
	double *buffer_edge_potential, temp_potential;
	unsigned int cur_edge, cur_index, i, j, loop_flag, step_size, carry_node_potential, total;
	vector < double> * cur_edge_potential;

	// first record to which edge each node is assigned 
	buffer_edge_potential = new double[dim_action * dim_action];	
	total = dim_action * dim_action;

	loop_flag = 0;
	for (cur_edge = 0; cur_edge < num_edge_max; cur_edge++)
	{		
		if(cur_edge < num_nodes)
		{
			loop_flag++;
			if(loop_flag == dim_horizontal)		// skip non-existent edges
			{
				loop_flag = 0;
				continue;
			}
		}

		cur_index = 0;		// used to traverse the 2D dim_action * dim_action table (flattened to a 1D array)
		temp_potential = exp(edge_fea_para[cur_edge][0]);		// now only one feature, so [0]

		for(i = 0; i < dim_action; i++)		// this loop is due to the indicator feature function
		{
			for(j = 0; j < i; j++)
			{
				buffer_edge_potential[cur_index++] = 1;
			}
			buffer_edge_potential[cur_index++] = temp_potential;
			for(j++; j < dim_action; j++)
			{
				buffer_edge_potential[cur_index++] = 1;
			}
		}
/*
		if(node_assign[cur_edge] >= 0)			// there is a node feature attached to this edge
		{
			carry_node_potential = (unsigned int) node_assign[cur_edge];		// which node?

			for(i = 0; i < dim_action; i++)
			{
				temp_potential = 0;
				for(j = 0; j < dim_observation; j++)
					temp_potential += node_fea_para[carry_node_potential][i][j] 
										* cur_observation[carry_node_potential][j];
				buffer_node_potential[i] = exp(temp_potential);			// temp_potential is the NN output
			}

			if(carry_node_potential == cur_edge)			// the node feature is on the first node
			{
				cur_index = 0;
				for (i = 0; i < dim_action; i++)
					for (j = 0; j < dim_action; j++)
						buffer_edge_potential[cur_index++] *= buffer_node_potential[i];
			}
			else											// the node feature is on the second node
			{
				for (i = 0; i < dim_action; i++)
				{
					cur_index = i;
					for (j = 0; j < dim_action; j++)
					{
						buffer_edge_potential[cur_index] *= buffer_node_potential[j];
						cur_index += dim_action;
					}
				}
			}
		}
*/
		// now do something on edge_potential_table_crf[] based on buffer_edge_potential[i]
		cur_edge_potential = edge_potential_table_crf[cur_edge];

		for (i = 0; i < total; i++)
		{
			(*cur_edge_potential)[i] = buffer_edge_potential[i];
		}
	}

	delete []buffer_edge_potential;
}


/*
 *	Update the potential table of the nodes, to be called in each iteration of RL.
 *	It does NOT reallocate the VECTOR space for node_potential_table_crf.  Just change the value in the vector.
 *	This function is very application specific.  For different tasks, one may want to re-write this function.
 *	Now it is written for 2D CRF with NN node features (linear)
 */
void update_node_potential_grid()
{
	double temp_potential;
	unsigned int cur_node, i, j;
	vector < double> * cur_node_potential;

	// first record to which edge each node is assigned	
	for(cur_node = 0; cur_node < num_nodes; cur_node ++)
	{
		cur_node_potential = node_potential_table_crf[cur_node];

		for(i = 0; i < dim_action; i++)
		{
			temp_potential = 0;

			for(j = 0; j < dim_observation; j++)
				temp_potential += node_fea_para[cur_node][i][j] * cur_observation[cur_node][j];
			(*cur_node_potential)[i] = exp(temp_potential);			// temp_potential is the NN output
		}
	}
}


/*************************************************************************************/
/*		End of updating potentials, for Grid										 */
/*************************************************************************************/



/*************************************************************************************/
/*		For testing, for Grid														 */
/*************************************************************************************/


/*
 *	Set the CRF edge potential for testing in the beginning. (run only once)
 *	vector < double> ** edge_potential_table_crf:
 *		The edge potential table.  [1]->: first index is the edge id, then get the pointer to the vector <double>.
 */
void set_edge_potential_for_test_1()
{
	double *buffer_edge_potential, temp_potential;
	unsigned int cur_edge, cur_index, i, j, loop_flag, step_size, carry_node_potential, total;
	vector < double> * cur_edge_potential;
	double large;

	// first record to which edge each node is assigned 
	buffer_edge_potential = new double[dim_action * dim_action];

	total = dim_action * dim_action;
	large = 1;

	loop_flag = 0;
	for (cur_edge = 0; cur_edge < num_edge_max; cur_edge++)
	{		
		if(cur_edge < num_nodes)
		{
			loop_flag++;
			if(loop_flag == dim_horizontal)		// skip non-existent features
			{
				loop_flag = 0;
				continue;
			}
		}

		cur_index = 0;		// used to traverse the 2D dim_action * dim_action table (flattened to a 1D array)

		if(cur_edge < num_nodes)		// horizontal edges
		{
			for(i = 0; i < dim_action; i++)
			{
				for(j = 0; j < (i + 1) % dim_action; j++)
					buffer_edge_potential[cur_index++] = small;
				buffer_edge_potential[cur_index++] = large;
				for(j++; j < dim_action; j++)
					buffer_edge_potential[cur_index++] = small;
			}
		}
		else							// vertical edges
		{
			for(i = 0; i < dim_action; i++)
			{
				for(j = 0; j < (i - 1) % dim_action; j++)
					buffer_edge_potential[cur_index++] = small;
				buffer_edge_potential[cur_index++] = large;
				for(j++; j < dim_action; j++)
					buffer_edge_potential[cur_index++] = small;
			}			
		}

/*		
		temp_potential = exp(edge_fea_para[cur_edge][0]);		// now only one feature, so [0]

		for(i = 0; i < dim_action; i++)		// this loop is due to the indicator feature function
		{
			for(j = 0; j < i; j++)
			{
				buffer_edge_potential[cur_index++] = 1;
			}
			buffer_edge_potential[cur_index++] = temp_potential;
			for(j++; j < dim_action; j++)
			{
				buffer_edge_potential[cur_index++] = 1;
			}
		}


		if(node_assign[cur_edge] >= 0)			// there is a node feature attached to this edge
		{
			carry_node_potential = (unsigned int) node_assign[cur_edge];		// which node?

			for(i = 0; i < dim_action; i++)
			{
				temp_potential = 0;
				for(j = 0; j < dim_observation; j++)
					temp_potential += node_fea_para[carry_node_potential][i][j] 
										* cur_observation[carry_node_potential][j];
				buffer_node_potential[i] = exp(temp_potential);			// temp_potential is the NN output
			}

			
			if(carry_node_potential == cur_edge)			// the node feature is on the first node
			{
				cur_index = 0;
				for (i = 0; i < dim_action; i++)
					for (j = 0; j < dim_action; j++)
						buffer_edge_potential[cur_index++] *= buffer_node_potential[i];
			}
			else											// the node feature is on the second node
			{
				for (i = 0; i < dim_action; i++)
				{
					cur_index = i;
					for (j = 0; j < dim_action; j++)
					{
						buffer_edge_potential[cur_index] *= buffer_node_potential[j];
						cur_index += dim_action;
					}
				}
			}
		}
*/
		// now do something on potential_table_crf[] based on buffer_edge_potential[i]

		cur_edge_potential = edge_potential_table_crf[cur_edge];

		for (i = 0; i < total; i++)
		{
			(*cur_edge_potential)[i] = buffer_edge_potential[i];
		}
	}

	delete []buffer_edge_potential;
}



/*
 *	Set the CRF node potential for testing in the beginning. (run only once)
 *	vector < double> ** node_potential_table_crf:
 *		The node potential table.  [1]->: first index is the node id, then get the pointer to the vector <double>.
 */
void set_node_potential_for_test_1()
{
	double temp_potential;
	unsigned int cur_node, i, j;
	vector < double> * cur_node_potential;

	for(cur_node = 0; cur_node < num_nodes; cur_node ++)
	{
		cur_node_potential = node_potential_table_crf[cur_node];

		for(i = 0; i < dim_action; i++)
		{
			temp_potential = 0;

			for(j = 0; j < dim_observation; j++)
				temp_potential += node_fea_para[cur_node][i][j] * cur_observation[cur_node][j];
			(*cur_node_potential)[i] = pow(10, temp_potential);			// temp_potential is the NN output
		}
	}
}


/*
 *	run at the beginning of RL, to set the node feature parameters for test1
 *	***node_fea_para:  current parameters for the node features, [1][2][3]
 *						first index: node id;	second index: action;	third index: input dimension
 */
void set_node_fea_para_for_test_1()
{
	unsigned int i, j, k;
	double *** cur_3, ** cur_2, * cur_1;

	cur_3 = node_fea_para;

	for(i = 0; i < num_nodes; i++, cur_3++)
	{
		cur_2 = *cur_3;

		for(j = 0; j < dim_action; j++, cur_2 ++)
		{
			cur_1 = *cur_2;
			for(k = 0; k < dim_observation; k++, cur_1++)
			{				
				*cur_1 = 0;				// this is actually node_fea_para[i][j][k]
			}

			assert(dim_observation == dim_action);
			*(*cur_2 + j) = weight;
		}
	}
}


/* 
 *	Set the current observation matrix for test1,   [1][2]
 *		1st index: node id;		2nd index: observation dimension index
 *	Will be run only once.
 */
void set_observation_for_test_1()
{
	unsigned int i, j;
	double **cur_2, *cur_1;

	cur_2 = cur_observation;

	for (i = 0; i < num_nodes; i++)
	{
		cur_1 = * (cur_2++);
		
		for (j = 0; j < dim_observation; j++)
		{
			*(cur_1 ++) = 0;			// actually cur_observation[i][j]
		}
	}

	cur_observation[0][0] = 1;
}


/*************************************************************************************/
/*		End of For testing, for Grid												 */
/*************************************************************************************/

















/*
 *	int		*time_edge_visit;			// how many times an edge has been covered, [1]
 *										//		index: edge id
 */
// void initialize_time_edge_visit()
// {
// 	unsigned int i, *cur_1;
// 
// 	time_edge_visit = new unsigned int [num_edge_max];
// 	cur_1 = time_edge_visit;
// 
// 	for(i = 0; i < num_edge_max; i++, cur_1++)
// 	{
// 		*(cur_1++) = 0;
// 	}
// }


/*
 *	int		*time_node_visit;			// how many times this node feature estimator has been evaluated, [1]
 *										//		index: node id
 */
// void initialize_time_node_visit()
// {
// 	unsigned int i, *cur_1;
// 
// 	time_node_visit = new unsigned int [num_nodes];
// 	cur_1 = time_node_visit;
// 
// 	for(i = 0; i < num_nodes; i++)
// 	{
// 		*(cur_1++) = 0;
// 	}
// }



/************************************************************************/
/*    Putting things together                                           */
/************************************************************************/


/*
 *	Set the CRF parameters for testing in the beginning. (run only once)
 */
void set_crf_parameters_for_test_1()
{
//	randomize_node_fea_para(1);

	set_observation_for_test_1();
	set_node_fea_para_for_test_1();
//	set_edge_fea_para_for_test_1();  NO need, because we will set edge potential directly.

	set_edge_potential_for_test_1();
	set_node_potential_for_test_1();
	clear_edge_margin_record();
}


/*
 *	Set up a CRF by specifying the following parameters:
 *	scalar:	dim_vertical, dim_horizontal, dim_observation, dim_action, num_nodes
 *	double * time_edge_visit, *time_node_visit
 *	double **:	edge_fea_para,      double ***:	node_fea_para
 *	This functions just specifies the FIXED parameters of the CRF, and allocate some space.  
 *	The update of the parameters in the run will be performed by update_crf_parameters().
 *	The real construction of CRF data structure is performed by set_up_model_from_string(stringstream & input)
 */
void initialize_crf_parameters(int m_dim_horizontal, int m_dim_vertical, int m_dim_observation, 
							   int m_dim_action, int m_num_edge_fea)
{
	unsigned int i, j;

	dim_horizontal = m_dim_horizontal;
	dim_vertical = m_dim_vertical;

	dim_observation = m_dim_observation;
	dim_action = m_dim_action;

	num_edge_fea = m_num_edge_fea;

//=======================================

	num_nodes = dim_horizontal * dim_vertical;
	num_edge_max = 2 * num_nodes - dim_horizontal;

	initialize_observation();
	initialize_edge_fea_para();
	initialize_node_fea_para();

	initialize_edge_potential();
	initialize_edge_margin_record();

	initialize_node_potential();
	initialize_node_margin_record();

	initialize_node_potential_aux();
	initialize_edge_potential_aux();

	sample_record = new unsigned int [num_nodes];
}


/*
 *	Randomize the CRF parameters in the beginning. (run only once)
 */
void randomize_parameters_grid(int argc, char** argv)
{
	double maxEdgeRand, maxNodeRand;

	maxEdgeRand = 1;
	getParaDouble(&maxEdgeRand, argc, argv, "-maxer", 0);

	maxNodeRand = 1;
	getParaDouble(&maxNodeRand, argc, argv, "-maxnr", 0);

	randomize_node_fea_para(maxNodeRand);
	randomize_edge_fea_para(maxEdgeRand);
	
	update_edge_potential_grid();
	update_node_potential_grid();

	clear_edge_margin_record();
	clear_node_margin_record();
	clear_node_grad();
	clear_edge_grad();
}


/*
 *	Update the CRF parameters in the run.
 */
void update_parameters_grid(double step_size)
{
	update_edge_fea_para(step_size);
	update_node_fea_para(step_size);

	update_edge_potential_grid();
	update_node_potential_grid();

	clear_edge_margin_record();
	clear_node_margin_record();
}


/************************************************************************/
/*    End of putting things together                                    */
/************************************************************************/






/************************************************************************/
/*   Getting statistics from the inference result						*/
/************************************************************************/


void print_edge_potential()
{
	double *buffer_edge_potential, *buffer_node_potential, temp_potential;
	unsigned int cur_edge, cur_index, i, j, loop_flag;
	vector < double> * temp_vec_potential;

	// first record to which edge each node is assigned 
	loop_flag = 0;
	for (cur_edge = 0; cur_edge < num_edge_max; cur_edge++)
	{		
		if(cur_edge < num_nodes)
		{
			loop_flag++;
			if(loop_flag == dim_horizontal)		// skip non-existent features
			{
				loop_flag = 0;
				continue;
			}
		}

		cur_index = 0;
		cout<<"== "<<cur_edge<<"==\n";
		
		for(i = 0; i < dim_action; i++)
		{
			for(j = 0; j < dim_action; j++)
			{
				cout<<(*edge_potential_table_crf[cur_edge])[cur_index++]<<"\t";
			}
			cout<<endl;
		}
	}
}


void cal_fea_expectations_grid()
{
	
	unsigned int i, j, k, cur_node, cur_edge;
	double *** cur_3, ** cur_2, * cur_1;
	
	//  First, nodes
	cur_3 = node_fea_exp;

	for(i = 0; i < num_nodes; i++)
	{
		cur_2 = *(cur_3++);

		for(j = 0; j < dim_action; j++)
		{
			cur_1 = *(cur_2++);

			for(k = 0; k < dim_observation; k++)
			{
				*(cur_1++) = node_margin_record[i][j] * cur_observation[i][k];
			}
		}
	}

	cur_2 = edge_fea_exp;
	cur_edge = 0;

	// Second, horizontal edges
	for(i = 0; i < dim_vertical; i++)
	{
		for(j = 0; j < dim_horizontal - 1; j++)
		{
			cur_1 = *(cur_2++);

			for(k = 0; k < num_edge_fea; k++)
			{
				*(cur_1++) = edge_margin_record[cur_edge][0][0] + edge_margin_record[cur_edge][1][1];
			}

			cur_edge++;
		}

		cur_edge++;
		cur_2++;
	}

	// Third, vertical edges
	for(i = dim_vertical * dim_horizontal; i < num_edge_max; i++)
	{
		cur_1 = *(cur_2++);

		for(k = 0; k < num_edge_fea; k++)
		{
			*(cur_1++) = edge_margin_record[cur_edge][0][0] + edge_margin_record[cur_edge][1][1];
		}
		cur_edge++;
	}	
}


/************************************************************************/
/*   End of getting statistics from the inference result				*/
/************************************************************************/


/*
 *	Test everything about the PGM
 *	For debug use only
 */
int test_PGM_string(int argc,  char * const argv[])
{

	stringstream input_string;
	double maxEdgeRand, maxNodeRand, maxObsRand, time_length;

	dim_horizontal = 3;
	getParaUnInt(&dim_horizontal, argc, argv, "-dh", 0);

	dim_vertical = 3;
	getParaUnInt(&dim_vertical, argc, argv, "-dv", 0);

	dim_observation = 2;
	getParaUnInt(&dim_observation, argc, argv, "-do", 0);

	dim_action = 2;
	getParaUnInt(&dim_action, argc, argv, "-da", 0);

	num_edge_fea = 1;
	getParaUnInt(&num_edge_fea, argc, argv, "-nef", 0);

	maxEdgeRand = 1;
	getParaDouble(&maxEdgeRand, argc, argv, "-maxer", 0);

	maxNodeRand = 1;
	getParaDouble(&maxNodeRand, argc, argv, "-maxnr", 0);

	maxObsRand = 1;
	getParaDouble(&maxObsRand, argc, argv, "-maxor", 0);

	time_length = 2;
	getParaDouble(&time_length, argc, argv, "-tl", 0);

	small = 1;
	getParaDouble(&small, argc, argv, "-sm", 0);

	weight = 1;
	getParaDouble(&weight, argc, argv, "-we", 0);

//	initialize_crf_parameters(int m_dim_horizontal, int m_dim_vertical, int m_dim_observation, 
//										int m_dim_action, int m_num_edge_fea)
	initialize_crf_parameters(dim_horizontal, dim_vertical, dim_observation, dim_action, dim_action);

	set_crf_parameters_for_test_1();
//	randomize_crf_parameters(maxObsRand, maxNodeRand, maxEdgeRand);

	prepare_string_input(input_string, (int)time_length, NULL);
	cout<<input_string.str();
	
	Experiment *main_experiment;
	
	main_experiment = new Experiment;

	if (!main_experiment->set_up_model_from_string(input_string))
	{
		cout << "A fatal error occurred while parsing file. Aborting..." << endl;
		return 1;
	}
	
//	print_edge_potential();
//	main_experiment->show_edges();
	main_experiment->get_inference();


	print_edge_marginals();
	cal_fea_expectations_grid();

//	main_experiment.export_results_to_file( argv[2]);
	
	delete main_experiment;
		
    return 0;
}


