/**
 * $Id: OLPomdp.cc 40 2007-01-19 06:02:40Z pakk $
 */

#include"PGBasics.hh"
#include<time.h>
#include"OLPomdp.hh"

namespace libpg {

/**
 * Top level piece of glue for Policy gradient, link the controller
 * and the simulator 
 * @param  Controller 
 * @param  Simulator (problem specific) 
 * @param  discount factor [0, 1]. If 1, trace must be reset
 * occasionally or have 0 mean reward.
 */
OLPomdp::OLPomdp(Controller* controller, Simulator* simulator, double discount, double stepSize) : RLAlg(controller, simulator, discount, stepSize) {};


/**
 * Do a fixed number of individual steps, aggregating the reward.  No
 * output is done. Could also be termed an "epoch" 
 * @param  number of steps to iterate for 
 * @param  should learning be turned on, or are we
 * just exploiting/evaluating the current policy?
 */
double OLPomdp::doSteps(Vector& totalRewards, int steps, bool learn) 
{
	int action_result;
	Vector action(simulator->getActionDim());
	Vector rewards(simulator->getRewardDim());
	Vector reinforcement(simulator->getRewardDim());

	totalRewards.clear();

	for (int s = 0; s < steps; s++) 
	{
		// Must do next line before calling getAction() since
		// getAction will accumulate log action gradients into the
		// trace directly for efficiency.
		if (learn && obs.getSteps() > 0) 
			controller->discountTrace();
		obs.setSteps(obs.getSteps()+1);
		simulator->getObservation(obs);
		controller->getAction(obs, action, learn);
		
		action_result = simulator->doAction(action);		

		// The last action is not obtained from controller's suggestion
		//	because the simulator might overrule the controller's decision.
		controller->buffer_last_action(simulator->get_last_action());

		if (action_result > 0)
			break; 

		s += action_result;		// if action_result = -1 (in sensor net case), then rewind.
		simulator->getReward(rewards);
		totalRewards += rewards;
		if (learn) 
		{
			reinforcement.assign(rewards - baseline);
			// [daa] Moved the check for 0 reward to the controller level
			//controller->instantStep(reinforcement, obs);
			controller->instantStep(reinforcement);
		}
	}

	totalRewards /= steps;
	return inner_prod(totalRewards, scalar_vector<double>(simulator->getRewardDim(), 1.0))/simulator->getRewardDim();
}

}
